import numpy as np
import copy
import scipy.spatial

from sklearn.mixture import GaussianMixture
from sklearn.cluster import SpectralClustering
from sklearn.metrics import pairwise

from scipy.spatial.distance import pdist, squareform
from scipy.spatial import distance as dist
from collections import Counter


def listify(obj):
    if isinstance(obj, str) or not hasattr(obj, "__iter__"):
        obj = [obj]
    return obj

class Metric:
    def __init__(self,data_variable='data',params=None,name=None,constrain_same=None,constrain_different=None):
        self.W = None
        if name is None:
            self.name = 'BaseMetric'
        else:
            self.name = name
        self.data_variable = data_variable
        if params is None:
            self.params = {}
        else:
            self.params = params
        
        if constrain_same is None:
            self.constrain_same = []
        else:
            self.constrain_same = constrain_same
            
        if constrain_different is None:
            self.constrain_different = []
        else:
            self.constrain_different = constrain_different
    
    
    def _get_X(self,dataset,transpose_var='sample'):
        X = dataset[listify(self.data_variable)].to_array().squeeze()
        X = X.transpose(transpose_var,...)#ensure that transpose_var is first dimension
        return X
    
    def to_dict(self):
        out = {'name':self.name,'data_variable':self.data_variable}
        out.update(self.params)
        return out
            
    def __repr__(self):
        return f'<Metric:{self.name}>'
    
    def copy(self):
        return copy.deepcopy(self)
    
    def __getitem__(self,index):
        return self.W[index]
    
    def __array__(self,dtype=None):
        return self.W.astype(dtype)
        
    def normalize1(self):
        W = self.W.copy()
        diag = np.diag(W)
        W = W/np.sqrt(np.outer(diag,diag))
        return W
    
    def normalize2(self):
        W = self.W.copy()
        diag = np.diag(W)
        W = (W - W.min())/(W.max()-W.min())
        return W 
    
    def apply_constraints(self):
        W = self.W.copy()
        for i,j in self.constrain_same:
            W[i,j] = 1.0
            W[j,i] = 1.0
            
        for i,j in self.constrain_different:
            W[i,j] = 0.0
            W[j,i] = 0.0
        return W
            
    
    def calculate(self,*args,**kwargs):
        raise NotImplementedError('Sub-classes must implement calculate!')
        
        
class Dummy(Metric):
    def __init__(self,data_variable='data',params=None,name=None,constrain_same=None,constrain_different=None):
        super().__init__(data_variable,params,name,constrain_same,constrain_different)
        if name is None:
            self.name = f'Dummy'
        
    def calculate(self,dataset,**params):
        n_samples = dataset.sizes['sample']
        self.W = np.identity(n_samples)
        return self
        
    
class Similarity(Metric):
    def __init__(self,data_variable='data',params=None,name=None,constrain_same=None,constrain_different=None):
        super().__init__(data_variable,params,name,constrain_same,constrain_different)
        if name is None:
            self.name = f'Similarity:{params["metric"]}'
        
    def calculate(self,dataset,**params):
        if params:
            self.params.update(params)
            
        X = self._get_X(dataset)
        if len(X.shape) == 1:
            print('reshaping input to similarity')
            X = X.expand_dims("feature",axis=1)
        self.W = pairwise.pairwise_kernels(
            X, 
            filter_params=True,  
            **self.params
        )
        
        return self
    
class Distance(Metric):
    def __init__(self,data_variable='data',params=None,name=None,constrain_same=None,constrain_different=None):
        super().__init__(data_variable,params,name,constrain_same,constrain_different)
        if name is None:
            self.name = f'Distance:{params["metric"]}'
        
    def calculate(self,dataset,**params):
        if params:
            self.params.update(params)
            
        X = dataset[listify(self.data_variable)].to_array().squeeze()
        self.W = pairwise.pairwise_distances(
            X, 
            **self.params
        )
        return self

class MultiMetric(Metric):
    def __init__(
        self, 
        metrics,
        data_variable='data',
        combine_by='prod',
        combine_by_powers=None,
        combine_by_coeffs=None,
        constrain_same=None,
        constrain_different=None,
        **params
    ): 
        super().__init__(data_variable,params,None,constrain_same,constrain_different)
        self.metrics=metrics
        self.update_name()
        
        self.combine_by=combine_by
        self.combine_by_powers = combine_by_powers
        self.combine_by_coeffs = combine_by_coeffs
        if self.combine_by=='prod':
            self.combine=self.prod
        elif self.combine_by=='sum':
            self.combine=self.sum
        else:
            raise ValueError('Combine by function not recognized. Must be "sum" or "prod"')
            
    def to_dict(self):
        out= {}
        for i,metric in enumerate(self.metrics):
            out[f'metric_{i}'] = metric.to_dict()
        return out
           
    def update_name(self):
        self.name=''
        for metric in self.metrics:
            self.name += metric.name+'-'
        self.name = self.name[:-1]#remove last '-'
    
    def prod(self,data_list):
        if self.combine_by_powers is not None:
            assert len(self.combine_by_powers)==len(data_list)
            powers = copy.deepcopy(self.combine_by_powers)
        else:
            powers = [1]*len(data_list)
        
        data_list = copy.deepcopy(data_list)
        
        #np methods use the __array__ accessor and return W
        value = np.power(data_list.pop(0),powers.pop(0))
        for data,power in zip(data_list,powers):
            value*=np.power(data,power)
        return value
    
    def sum(self,data_list):
        if self.combine_by_powers is not None:
            assert len(self.combine_by_coeffs)==len(data_list)
            coeffs = copy.deepcopy(self.combine_by_coeffs)
        else:
            coeffs = [1]*len(data_list)
        
        data_list = copy.deepcopy(data_list)
        
        #np methods use the __array__ accessor and return W
        value = np.multiply(data_list.pop(0),coeffs.pop(0))
        for data,coeff in zip(data_list,coeffs):
            value += np.multiply(data,coeff)
            
        return value
        
    def calculate(self,dataset):
        W_list = []
        for metric in self.metrics:
            W_list.append(metric.calculate(dataset))
        self.W_list = W_list
        self.W = self.combine(W_list)
        self.W = self.normalize1()
        self.W = self.apply_constraints()
        return self
    
class Delaunay(Similarity):
    def __init__(self,data_variable='data',params=None):
        super().__init__(data_variable,params)
        self.name = 'Delaunay'
        
    def calculate(self,X):
        """
        Computes the Delaunay triangulation of the given points
        :param x: array of shape (num_nodes, 2)
        :return: the computed adjacency matrix
        """
        tri = scipy.spatial.Delaunay(X)
        edges_explicit = np.concatenate((tri.vertices[:, :2],
                                         tri.vertices[:, 1:],
                                         tri.vertices[:, ::2]), axis=0)
        adj = np.zeros((x.shape[0], x.shape[0]))
        adj[edges_explicit[:, 0], edges_explicit[:, 1]] = 1.
        self.W = np.clip(adj + adj.T, 0, 1) 
        return self
    
    



import threading
from typing import Any, List


class NoContextException(Exception):
    pass


class PipelineContext:
    """Base class that provides context manager functionality for pipeline operations.
    
    This class implements a thread-local context stack pattern that allows pipeline
    operations to be associated with their parent pipeline. When a Pipeline instance
    is used as a context manager (with statement), it pushes itself onto the context
    stack, making it available to operations created within that context.
    
    This pattern enables a more intuitive API where operations can automatically
    associate with the currently active pipeline without explicit references.
    
    Examples
    --------
    >>> with Pipeline(name="my_pipeline") as pipe:
    ...     # Operations created here can access the pipeline via get_context()
    ...     op = SomeOperation(input_var="data", output_var="result")
    ...     # op can now reference the pipeline context
    
    Notes
    -----
    This implementation uses thread-local storage to ensure thread safety when
    multiple pipelines are being constructed simultaneously in different threads.
    
    See Also
    --------
    Pipeline : Main container class that inherits this context functionality
    """

    contexts = threading.local()

    def __enter__(self):
        type(self).get_contexts().append(self)
        return self

    def __exit__(self, typ, value, traceback):
        type(self).get_contexts().pop()

    @classmethod
    def get_contexts(cls) -> List:
        if not hasattr(cls.contexts, "stack"):
            cls.contexts.stack = []
        return cls.contexts.stack

    @classmethod
    def get_context(cls) -> Any:
        """Return the deepest context on the stack."""
        try:
            return cls.get_contexts()[-1]
        except IndexError:
            raise NoContextException("No context on context stack")



import numpy as np
import pandas as pd
import xarray as xr
from math import sqrt

import plotly.graph_objects as go
import plotly.express as px


import ipywidgets
from sklearn.preprocessing import OrdinalEncoder

class ALPlot:
    def __init__(self,saxs_data,composition_data,possible_phase_labels):
        self.data_view = DiffractionLabelerView()
        self.data_model = DiffractionLabelerModel(saxs_data,composition_data,possible_phase_labels)
        self.data_index = 0
        
    def next_button_callback(self,click):
        self.data_index+=1
        self.data_view.current_index.value = self.data_index
        self.update_plot()
        
    def prev_button_callback(self,click):
        self.data_index-=1
        self.data_view.current_index.value = self.data_index
        self.update_plot()
        
    def goto_callback(self,click):
        index = self.data_view.current_index.value
        self.data_index=index
        self.update_plot()
    
    def ternary_click_callback(self,figure,location,click):
        index = location.point_inds[0]
        self.data_index = index
        self.data_view.current_index.value = self.data_index
        self.update_plot()
        
    def update_plot(self):
        saxs_data = self.data_model.saxs_data[self.data_index]
        composition_data = self.data_model.composition_data[self.data_index]
        self.data_view.update_plot(x=saxs_data.q.values,y=saxs_data.values,composition=composition_data)
        self.data_view.current_label.value = self.data_model.phase_labels[self.data_index]
        
    def draw_peaks(self,peaks):
        self.data_view.output.clear_output()
        with self.data_view.output:
            for i,(vl,(spacing,peak_loc)) in enumerate(zip(self.data_view.fig1.layout.shapes,peaks.items())):
                print('{}q* = {}'.format(spacing,peak_loc))
                vl.update({ 
                    'x0':peak_loc,
                    'x1':peak_loc,
                    'visible':True
                })
                    
            for i,vl, in enumerate(self.data_view.fig1.layout.shapes):
                if i>len(peaks)-1:
                    vl.visible=False
        
    def change_qstar_callback(self,figure,location,click):
        model = self.data_view.phase_dropdown.value
        n_orders = self.data_view.n_orders.value
        peaks = self.data_model.get_peaks(model,qstar=location.xs[0],max_order=n_orders)
        self.draw_peaks(peaks)
            
    def change_model_callback(self,data):
        model = self.data_view.phase_dropdown.value
        n_orders = self.data_view.n_orders.value
        peaks = self.data_model.get_peaks(model,max_order=n_orders)
        self.draw_peaks(peaks)
        
    def change_norder_callback(self,data):
        model = self.data_view.phase_dropdown.value
        n_orders = self.data_view.n_orders.value
        peaks = self.data_model.get_peaks(model,max_order=n_orders)
        self.draw_peaks(peaks)
        
    def label(self,label):
        self.data_model.phase_labels[self.data_index] = label
        self.data_view.update_ternary_colors(self.data_model.ordinal_phase_labels())
        self.next_button_callback(None)
        
    def run(self):
        saxs_data = self.data_model.saxs_data[self.data_index]
        composition_data = self.data_model.composition_data[self.data_index]
        widget = self.data_view.run(
            x = saxs_data.q.values,
            y = saxs_data.values,
            all_compositions = self.data_model.composition_data,
            composition = composition_data,
            models = list(self.data_model.models.keys()),
            possible_phase_labels=[],
        )
        
        self.data_view.intensity.on_click(self.change_qstar_callback)
        self.data_view.phase_dropdown.observe(self.change_model_callback)
        self.data_view.current_label.value = self.data_model.phase_labels[self.data_index]
        self.data_view.current_index.value = str(self.data_index)
        self.data_view.bnext.on_click(self.next_button_callback)
        self.data_view.bprev.on_click(self.prev_button_callback)
        self.data_view.bgoto.on_click(self.goto_callback)
        self.data_view.all_ternary.on_click(self.ternary_click_callback)
        self.data_view.n_orders.observe(self.change_norder_callback)
        
        self.data_view.b0.on_click(lambda click: self.label('A'))
        self.data_view.b1.on_click(lambda click: self.label('B'))
        self.data_view.b2.on_click(lambda click: self.label('C'))
        self.data_view.b3.on_click(lambda click: self.label('D'))
        self.data_view.b4.on_click(lambda click: self.label('E'))
        
        return widget
    
##################
### Data Model ###
##################

class DiffractionLabelerModel:
    def __init__(self,saxs_data,composition_data,possible_phase_labels):
        self.saxs_data = saxs_data
        self.composition_data = composition_data
        self.phase_labels = ['Unlabeled']*len(composition_data)
        self.possible_phase_labels = possible_phase_labels
            
        self.qstar = 0.02
        self.init_models()
        
    def ordinal_phase_labels(self):
        enc = OrdinalEncoder()
        return enc.fit_transform(np.asarray(self.phase_labels)[:,np.newaxis]).flatten()
        
    def init_models(self):
        self.models = {}
        self.models['q*'] = { 
            'labels':["1"],
            'spacings':[1]
        }
        self.models['LAM'] = { 
            'labels':["1","2","3","4","5","6","7","8","9","10","11","12","13","14","15","16","17","18","19","20"],
            'spacings':[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20]
        }
        self.models['HCP CYL'] = { 
            'labels': ["1"," sqrt(3)"," 2"," sqrt(7)"," 3"," sqrt(12)"," sqrt(13)"," 4"," sqrt(19)"," sqrt(21)"," 5"," sqrt(27)"," sqrt(28)"," sqrt(31)"," 6"," sqrt(37)"," sqrt(39)"," sqrt(43)"," sqrt(48)","7"],
            'spacings': [1, sqrt(3), 2, sqrt(7), 3, sqrt(12), sqrt(13), 4, sqrt(19), sqrt(21), 5, sqrt(27), sqrt(28), sqrt(31), 6, sqrt(37), sqrt(39), sqrt(43), sqrt(48), 7]
        }
        self.models['SC'] = { 
            'labels': ["1","sqrt(2)","sqrt(3)","2","sqrt(5)","sqrt(6)","sqrt(8)","3"],
            'spacings': [1,sqrt(2),sqrt(3),2,sqrt(5),sqrt(6),sqrt(8),3]
        }
        self.models['BCC'] = { 
            'labels':["1","sqrt(2)","sqrt(3)","2","sqrt(5)","sqrt(6)","sqrt(7)","sqrt(8)","3"],
            'spacings':[1,sqrt(2),sqrt(3),2,sqrt(5),sqrt(6),sqrt(7),sqrt(8),3],
        }
        self.models['FCC'] = { 
            'labels':["sqrt(3)","2","sqrt(8)","sqrt(11)","sqrt(12)","4","sqrt(19)"],
            'spacings':[1,2/sqrt(3),sqrt(8)/sqrt(3),sqrt(11)/sqrt(3),sqrt(12)/sqrt(3),4/sqrt(3),sqrt(19)/sqrt(3)],
        }
        self.models['SPH'] = { 
            'labels':["1","2/sqrt(3)","2*sqrt(2)/sqrt(3)","2"],
            'spacings':[1,2/sqrt(3),2*sqrt(2)/sqrt(3),2],
        }
        self.models['HCP SPH'] = { 
            'labels':["sqrt(32)","6","sqrt(41)","sqrt(68)","sqrt(96)","sqrt(113)"],
            'spacings':[1,6/sqrt(32),sqrt(41)/sqrt(32),sqrt(68)/sqrt(32),sqrt(96)/sqrt(32),sqrt(113)/sqrt(32)],
        }
        self.models['DD'] = { 
            'labels':["sqrt(2)"," sqrt(3)"," 2"," sqrt(6)"," sqrt(8)"," 3"," sqrt(10)"," sqrt(11)"],
            'spacings':[1, sqrt(3)/sqrt(2), 2/sqrt(2), sqrt(6)/sqrt(2), sqrt(8)/sqrt(2), 3/sqrt(2), sqrt(10)/sqrt(2), sqrt(11)/sqrt(2)],
        }
        self.models['GYR'] = { 
            'labels':["sqrt(6)","sqrt(8)","sqrt(14)","sqrt(16)","sqrt(20)","sqrt(22)","sqrt(24)","sqrt(26)","sqrt(30)","sqrt(32)","sqrt(34)","sqrt(38)","sqrt(40)","sqrt(42)","sqrt(46)","sqrt(48)","sqrt(50)","sqrt(52)","sqrt(56)","sqrt(62)","sqrt(64)","sqrt(66)","sqrt(68)","sqrt(70)","sqrt(72)","sqrt(74)","sqrt(78)","sqrt(80)","sqrt(84)","sqrt(86)","sqrt(88)","sqrt(90)"],
            'spacings':[6,8,14,16,20,22,24,26,30,32,34,38,40,42,46,48,50,52,56,62,64,66,68,70,72,74,78,80,84,86,88,90],
        }
        self.models['GYR']['spacings'] = [sqrt(i)/sqrt(6) for i in self.models['GYR']['spacings']]
        
    def get_peaks(self,model,qstar=None,max_order=4):
        if qstar is None:
            qstar=self.qstar
        else:
            self.qstar=qstar
            
        peaks = {}
        for i,(spacing,label) in enumerate(zip(self.models[model]['spacings'],self.models[model]['labels'])):
            if i>=max_order:
                break
            peaks[label] = spacing*qstar
        return peaks
        
        
        
#################
### Data View ###
#################
    
class DiffractionLabelerView:
    def __init__(self):
        self.intensity = None
        self.nverts = 8
        
    def update_plot(self,x,y,composition):
        self.intensity.update({'x':x,'y':y})
        self.ternary.update({
            'a':(composition[0],),
            'b':(composition[1],), 
            'c':(composition[2],)
        })
        
    def remove_vertical_lines(self):
        self.fig1.layout['shapes'] = []
        
    def add_vertical_line(self,x,y0=0,y1=128,row=1,col=1,line_kw=None):
        if line_kw is None:
            line_kw=dict(color='red',dash='dot',width=0.3)

        self.fig1.add_shape(
            name='vertical',
            xref='x',
            yref='paper',
            x0=x, x1=x, y0=0, y1=1,
            line=line_kw,
        )
        
    def update_ternary_colors(self,colors):
        self.all_ternary.marker['color'] = colors
        
    def run(self,x,y,all_compositions,composition,models,possible_phase_labels):
        self.fig1 = go.FigureWidget(
            go.Scatter(x=x,y=y)
        )
        self.intensity = self.fig1.data[0]
        self.fig1.update_yaxes(type='log')
        self.fig1.update_xaxes(type='log')
        #self.fig1.update_xaxes({'range':(0,0.1)})
        self.fig1.update_layout(height=300,width=400,margin=dict(t=10,b=10,l=10,r=0))
        
        for i in range(self.nverts):
            self.add_vertical_line(0.02*i)
            self.fig1.layout.shapes[i].visible = False
        
        self.fig2 = go.FigureWidget([ 
            go.Scatterternary( 
                a = all_compositions[:,0], 
                b = all_compositions[:,1], 
                c = all_compositions[:,2], 
                mode = 'markers', 
                marker={
                    'color':['black']*len(x),
                    'colorscale':px.colors.qualitative.Prism,
                },
                customdata = list(range(len(all_compositions))),
                opacity=1.0,
                showlegend=False,
            ),
            go.Scatterternary( 
                a = (composition[0],), 
                b = (composition[1],), 
                c = (composition[2],), 
                mode = 'markers', 
                showlegend=False,
                marker={
                    'color':'red',
                    'symbol':'hexagon-open',
                    'size':10,
                },
            ),
            ]
        )
        self.all_ternary = self.fig2.data[0]
        self.ternary = self.fig2.data[1]
        self.fig2.update_layout(height=300,width=500,margin=dict(t=25,b=35,l=10))
        
        plot_box = ipywidgets.HBox([self.fig1,self.fig2])
        
        self.b0 = ipywidgets.Button(description='A')
        self.b1 = ipywidgets.Button(description='B')
        self.b2 = ipywidgets.Button(description='C')
        self.b3 = ipywidgets.Button(description='D')
        self.b4 = ipywidgets.Button(description='E')
        self.br = ipywidgets.Button(description='Reset')
        self.bprev = ipywidgets.Button(description='Prev')
        self.bnext = ipywidgets.Button(description='Next')
        self.bgoto = ipywidgets.Button(description='GoTo')
        self.current_index = ipywidgets.IntText(description="Data Index:",value=0)
        
        self.n_orders = ipywidgets.BoundedIntText(description='n_orders',min=1,max=8,value=4)
        self.phase_dropdown = ipywidgets.Dropdown(
            options=models
        )
        self.current_label_label = ipywidgets.Label('Current Label:')
        self.current_label = ipywidgets.Label('')
        
        phase_box = ipywidgets.HBox([
            self.phase_dropdown,
            self.n_orders,
            self.current_label_label,
            self.current_label,
        ])
        
        buttons_hbox1 = ipywidgets.HBox([self.b0,self.b1,self.b2,self.b3,self.b4,self.br])
        buttons_hbox2 = ipywidgets.HBox([
            self.current_index,
            self.bgoto,
            self.bprev,
            self.bnext,
        ])
        self.output = ipywidgets.Output()
        vbox = ipywidgets.VBox([
            buttons_hbox1,
            buttons_hbox2,
            phase_box,
            plot_box,
            self.output
        ])
        
        return vbox
    
        
    


from AFL.automation.APIServer.Client import Client
from AFL.automation.prepare.OT2Client import OT2Client
from AFL.automation.shared.utilities import listify
from AFL.automation.APIServer.Driver import Driver
from AFL.automation.shared.Serialize import serialize,deserialize
from AFL.agent.AgentClient import AgentClient
from AFL.automation.shared.units import units

import pandas as pd
import numpy as np
import xarray as xr

from math import ceil,sqrt
import json
import time
import requests
import shutil
import datetime
import traceback
import pathlib
import uuid

import AFL.automation.prepare
import shutil
import h5py


class SAXS_AL_SampleDriver(Driver):
    defaults={}
    defaults['snapshot_directory'] = '/home/nistoroboto'
    defaults['data_path'] = '/nfs/aux/chess/reduced_data/cycles/2022-1/id3b/beaucage-2324-D/analysis/'
    defaults['data_manifest_file'] = '/nfs/aux/chess/reduced_data/cycles/2022-1/id3b/beaucage-2324-D/analysis/data_manifest.csv'
    defaults['data_tag'] = 'default'
    def __init__(self,
            load_url,
            prep_url,
            sas_url,
            agent_url,
            camera_urls = None,
            snapshot_directory =None,
            overrides=None, 
            dummy_mode=False,
            ):

        Driver.__init__(self,name='SAXS_AL_SampleDriver',defaults=self.gather_defaults(),overrides=overrides)

        if not (len(load_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on load_url')

        if not (len(prep_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on prep_url')

        if not (len(agent_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on agent_url')

        self.app = None
        self.name = 'SAXS_AL_SampleDriver'

        self.dummy_mode= dummy_mode
        if not self.dummy_mode:
            #prepare samples
            self.prep_url = prep_url
            self.prep_client = OT2Client(prep_url.split(':')[0],port=prep_url.split(':')[1])
            self.prep_client.login('SampleServer_PrepClient')
            self.prep_client.debug(False)
            
            #load samples
            self.load_client = Client(load_url.split(':')[0],port=load_url.split(':')[1])
            self.load_client.login('SampleServer_LoadClient')
            self.load_client.debug(False)
    # 
            #load samples
            self.sas_url = sas_url
            self.sas_client = Client(sas_url.split(':')[0],port=sas_url.split(':')[1])
            self.sas_client.login('SampleServer_SASClient')
            self.sas_client.debug(False)
        else:
            self.sas_url = sas_url
            self.prep_url = prep_url
        
        #agent
        self.agent_url = agent_url
        self.agent_client = AgentClient(agent_url.split(':')[0],port=agent_url.split(':')[1])
        self.agent_client.login('SampleServer_AgentClient')
        self.agent_client.debug(False)

        self.camera_urls = camera_urls

        if snapshot_directory is not None:
            self.config['snapshot_directory'] = snapshot_directory

        self.rinse_uuid = None
        self.prep_uuid = None
        self.catch_uuid = None
        self.agent_uuid = None

        self.status_str = 'Fresh Server!'
        self.wait_time = 30.0 #seconds
        
        
        self.catch_protocol=None
        self.AL_status_str = ''
        self.data_manifest  = None
        self.AL_components = None
        
        self.reset_deck()
       
    def reset_deck(self):
        self.deck = AFL.automation.prepare.Deck()
        
    def add_container(self,name,slot):
        self.deck.add_container(name,slot)
        
    def add_catch(self,name,slot):
        self.deck.add_catch(name,slot)
        self.catch_loc=f"{slot}A1"
    
    def add_pipette(self,name,mount,tipracks):
        self.deck.add_pipette(name,mount,tipracks=tipracks)
        
    def send_deck_config(self,home=True):
        self.deck.init_remote_connection(
            self.prep_url.split(":")[0],
            home=home
        )
        self.deck.send_deck_config()
        
    def add_stock(self,stock_dict,loc):
        soln = AFL.automation.prepare.Solution.from_dict(stock_dict)
        self.deck.add_stock(soln,loc)
       
    def status(self):
        status = []
        status.append(f'Snapshots: {self.config["snapshot_directory"]}')
        status.append(f'Cameras: {self.camera_urls}')
        status.append(f'SAS: {self.sas_url}')
        status.append(f'Sample Wait Time: {self.wait_time}')
        status.append(f'AL_components: {self.AL_components}')
        status.append(f'{len(self.deck.stocks)} stocks loaded!')
        status.append(self.status_str)
        status.append(self.AL_status_str)
        return status
    
    def update_status(self,value):
        self.status_str = value
        self.app.logger.info(value)

    def take_snapshot(self,prefix):
        now = datetime.datetime.now().strftime('%y%m%d-%H:%M:%S')
        for i,cam_url in enumerate(self.camera_urls):
            fname = self.config['snapshot_directory'] + '/' 
            fname += prefix
            fname += f'-{i}-'
            fname += now
            fname += '.jpg'

            try:
                r = requests.get(cam_url,stream=True)
                if r.status_code == 200:
                    with open(fname,'wb') as f:
                        r.raw.decode_content=True
                        shutil.copyfileobj(r.raw,f)
            except Exception as error:
                output_str  = f'take_snapshot failed with error: {error.__repr__()}\n\n'+traceback.format_exc()+'\n\n'
                self.app.logger.warning(output_str)

    def measure(self,sample):
        exposure = sample.get('exposure',None)

        sas_uuid = self.sas_client.enqueue(task_name='expose',name=sample['name'],block=True,exposure=exposure)
        self.sas_client.wait(sas_uuid)

        return sas_uuid

    #def execute(self,**kwargs):
    #    if self.app is not None:
    #        self.app.logger.debug(f'Executing task {kwargs}')
    #
    #    if kwargs['task_name']=='sample':
    #        self.process_sample(kwargs)
    #    elif kwargs['task_name']=='measure':
    #        self.measure(kwargs)
    #    elif kwargs['task_name']=='take_snapshot':
    #        self.take_snapshot(kwargs['prefix'])
    #    elif kwargs['task_name']=='set_snapshot_directory':
    #        self.config['snapshot_directory'] = kwargs['snapshot_directory']
    #    elif kwargs['task_name']=='set_wait_time':
    #        self.wait_time = kwargs['wait_time']
    #    else:
    #        raise ValueError(f'Task_name not recognized: {kwargs["task_name"]}')

    def sample(self,sample):
        return self.process_sample(sample)
        
    def process_sample(self,sample):
        name = sample['name']

        targets = set()
        for task in sample['prep_protocol']:
            if 'target' in task['source'].lower():
                targets.add(task['source'])
            if 'target' in task['dest'].lower():
                targets.add(task['dest'])

        for task in sample['catch_protocol']:
            if 'target' in task['source'].lower():
                targets.add(task['source'])
            if 'target' in task['dest'].lower():
                targets.add(task['dest'])

        target_map = {}
        for t in targets:
            prep_target = self.prep_client.enqueue(task_name='get_prep_target',interactive=True)['return_val']
            target_map[t] = prep_target

        for task in sample['prep_protocol']:
            #if the well isn't in the map, just use the well
            task['source'] = target_map.get(task['source'],task['source'])
            task['dest'] = target_map.get(task['dest'],task['dest'])
            self.prep_uuid = self.prep_client.transfer(**task)
 
        if self.rinse_uuid is not None:
            self.update_status(f'Waiting for rinse...')
            self.load_client.wait(self.rinse_uuid)
            self.update_status(f'Rinse done!')

        self.update_status(f'Cell is clean, measuring empty cell scattering...')
        empty = {}
        empty['name'] = 'MT-'+sample['name']
        empty['exposure'] = sample['exposure']
        empty['wiggle'] = False
        self.sas_uuid = self.measure(empty)
        self.sas_client.wait(self.sas_uuid)
            
        if self.prep_uuid is not None: 
            self.prep_client.wait(self.prep_uuid)
            self.take_snapshot(prefix = f'02-after-prep-{name}')
        
        self.update_status(f'Queueing sample {name} load into syringe loader')
        for task in sample['catch_protocol']:
            #if the well isn't in the map, just use the well
            task['source'] = target_map.get(task['source'],task['source'])
            task['dest'] = target_map.get(task['dest'],task['dest'])
            self.catch_uuid = self.prep_client.transfer(**task)
        
        if self.catch_uuid is not None:
            self.update_status(f'Waiting for sample prep/catch of {name} to finish: {self.catch_uuid}')
            self.prep_client.wait(self.catch_uuid)
            self.take_snapshot(prefix = f'03-after-catch-{name}')
        
        
        self.load_uuid = self.load_client.enqueue(task_name='loadSample',sampleVolume=sample['volume'])
        self.update_status(f'Loading sample into cell: {self.load_uuid}')
        self.load_client.wait(self.load_uuid)
        self.take_snapshot(prefix = f'05-after-load-{name}')
        
        self.update_status(f'Sample is loaded, asking the instrument for exposure...')
        self.sas_uuid = self.measure(sample)

        self.update_status(f'Cleaning up sample {name}...')
        self.rinse_uuid = self.load_client.enqueue(task_name='rinseCell')

        self.update_status(f'Waiting for instrument to measure scattering of {name} with UUID {self.sas_uuid}...')
        self.sas_client.wait(self.sas_uuid)
        self.take_snapshot(prefix = f'06-after-measure-{name}')
            
        self.update_status(f'All done for {name}!')


    def process_sample_outoforder(self,sample):
        name = sample['name']

        targets = set()
        for task in sample['prep_protocol']:
            if 'target' in task['source'].lower():
                targets.add(task['source'])
            if 'target' in task['dest'].lower():
                targets.add(task['dest'])

        for task in sample['catch_protocol']:
            if 'target' in task['source'].lower():
                targets.add(task['source'])
            if 'target' in task['dest'].lower():
                targets.add(task['dest'])

        target_map = {}
        for t in targets:
            prep_target = self.prep_client.enqueue(task_name='get_prep_target',interactive=True)['return_val']
            target_map[t] = prep_target

        for task in sample['prep_protocol']:
            #if the well isn't in the map, just use the well
            task['source'] = target_map.get(task['source'],task['source'])
            task['dest'] = target_map.get(task['dest'],task['dest'])
            self.prep_uuid = self.prep_client.transfer(**task)
        
        if self.load_uuid is not None:
            self.load_client.wait(self.load_uuid)
            last_name = self.last_sample['name']
            self.take_snapshot(prefix = f'05-after-load-{last_name}')

            self.update_status(f'Sample is loaded, asking the instrument for exposure...')
            self.sas_uuid = self.measure(self.last_sample)#self.measure blocks...
            self.take_snapshot(prefix = f'06-after-measure-{last_name}')
            
            self.update_status(f'Cleaning up sample {last_name}...')
            self.rinse_uuid = self.load_client.enqueue(task_name='rinseCell')
            self.load_client.wait(self.rinse_uuid)

            self.update_status(f'Waiting for rinse...')
            self.load_client.wait(self.rinse_uuid)
            self.update_status(f'Rinse done!')
            
        self.update_status(f'Queueing sample {name} load into syringe loader')
        for task in sample['catch_protocol']:
            #if the well isn't in the map, just use the well
            task['source'] = target_map.get(task['source'],task['source'])
            task['dest'] = target_map.get(task['dest'],task['dest'])
            self.catch_uuid = self.prep_client.transfer(**task)
    
    @Driver.unqueued()
    def stop_active_learning(self):
        self.stop_AL = True
        
    def set_catch_protocol(self,**kwargs): 
        self.catch_protocol = AFL.automation.prepare.PipetteAction(**kwargs)
    
    def active_learning_loop(self,**kwargs):
        self.AL_components = kwargs['AL_components']
        sample_volume = kwargs['sample_volume']
        exposure = kwargs['exposure']
        mix_order = kwargs['mix_order']
        custom_stock_settings = kwargs['custom_stock_settings']
        data_manifest_path = pathlib.Path(self.config['data_manifest_file'])
        data_path = pathlib.Path(self.config['data_path'])
        
        if self.dummy_mode:
            raise ValueError('Dummy mode no longer supported')
            from AFL.automation.agent import PhaseMap
            df_measurements = pd.read_csv('/Users/tbm/projects/2001-NistoRoboto/2022-02-13-CodeDev/measurements.csv')
            
        self.stop_AL = False
        while not self.stop_AL:
            self.app.logger.info(f'Starting new AL loop')
            
            #get prediction of next step
            #next_sample = self.agent_client.get_next_sample_queued()
            next_sample = self.agent_client.get('next_sample')
            self.app.logger.info(f'Preparing to make next sample: {next_sample}')
            
            # if self.dummy_mode:
            #     print(f'Pulling from {df_measurements.shape}')
            #     xy_all = PhaseMap.ternary2cart(df_measurements[self.AL_components].values)
            #     xy_next = PhaseMap.ternary2cart(next_sample[self.AL_components])
            #     dist = np.sqrt(np.sum(np.square(xy_all-xy_next),axis=1))
            #     argmin = np.argmin(dist)
            #     next_sample_full = df_measurements.iloc[argmin]
            #     next_sample = next_sample_full[self.AL_components].to_frame().T
            #     df_measurements = df_measurements.drop(argmin).reset_index(drop=True)
            #     print('\n\n\n')
            #     print(next_sample_full)
            #     print(df_measurements)
            #     print('\n\n\n')
                
            #make target object
            target = AFL.automation.prepare.Solution('target',list(next_sample.columns.values))
            # target = AFL.automation.prepare.Solution('target',list(next_sample.columns.values) + ['NaCl'])
            target.mass_fraction = next_sample_dict
            target.volume = sample_volume*units('ul')
            # target.concentration = {'NaCl':110*units('mg/ml')}
            # target.concentration = {'F127':100*units('mg/ml')}
            # self.app.logger.info('Setting fixed F127 concentration of 100 mg/ml')
                
            self.deck.reset_targets()
            self.deck.add_target(target,name='target')
            self.deck.make_sample_series(reset_sample_series=True)
            self.deck.validate_sample_series(tolerance=0.15)
            self.deck.make_protocol(only_validated=False)
            self.fix_protocol_order(mix_order,custom_stock_settings)
            sample,validated = self.deck.sample_series[0]
            self.app.logger.info(self.deck.validation_report)
            
            if validated:
                self.app.logger.info(f'Validation PASSED')
                self.AL_status_str = 'Last sample validation PASSED'
            else:
                self.app.logger.info(f'Validation FAILED')
                self.AL_status_str = 'Last sample validation FAILED'
            self.app.logger.info(f'Making next sample with mass fraction: {sample.target_check.mass_fraction}')
            
            self.catch_protocol.source = sample.target_loc
            
            if self.dummy_mode:
                raise ValueError('Dummy mode no longer supported')
                # sample_name = next_sample_full['fname'].replace('_r1d','')
                # sample.target_check.mass_fraction
                # sample.target_check.mass_fraction = next_sample_dict
                # sample.target_check.volume = sample_volume*units('ul')
                # print('SAMPLE_NAME:',sample_name)
                # time.sleep(5)
            else:
                sample_uuid = str(uuid.uuid4())[-8:]
                sample_name = f'AL_{self.config["data_tag"]}_{sample_uuid}'
                self.process_sample(
                        dict(
                            name=sample_name,
                            prep_protocol = sample.emit_protocol(),
                            catch_protocol =[self.catch_protocol.emit_protocol()],
                            volume = self.catch_protocol.volume/1000.0,
                            exposure = exposure
                    )
                )
        

            # CHECK TRANMISSION OF LAST SAMPLE
            # XXX Need to update based on how files will be read
            file_path = data_path / (sample_name+'.txt')
            with h5py.File(h5_path,'r') as h5:
                transmission = h5['entry/sample/transmission'][()]

            if transmission>0.9:
                self.update_status(f'Last sample missed! (Transmission={transmission})')
                self.app.logger.info('Dropping this sample from AL and hoping the next one hits...')
                continue
            else:
                self.update_status(f'Last Sample success! (Transmission={transmission})')
            
            # update manifest
            if data_manifest_path.exists():
                self.data_manifest = pd.read_csv(data_manifest_path)
            else:
                self.data_manifest = pd.DataFrame(columns=['fname','label',*self.AL_components])
            
            row = {}
            row['fname'] = data_path/(sample_name+'_chosen_r1d.csv')
            row['label'] = -1
            total = 0
            for component in self.AL_components:
                m = sample.target_check.mass_fraction[component].magnitude
                row[component] = m
                total+=m
            for component in self.AL_components:
                row[component] = 100.0*(row[component]/total)
            
            self.data_manifest = self.data_manifest.append(row,ignore_index=True)
            self.data_manifest.to_csv(data_manifest_path,index=False)
            
            # trigger AL
            self.agent_uuid = self.agent_client.enqueue(task_name='predict')
            
            # wait for AL
            self.app.logger.info(f'Waiting for agent...')
            self.agent_client.wait(self.agent_uuid)
            
            
    def fix_protocol_order(self,mix_order,custom_stock_settings):
        mix_order = [self.deck.get_stock(i) for i in mix_order]
        mix_order_map = {loc:new_index for new_index,(stock,loc) in enumerate(mix_order)}
        for sample,validated in self.deck.sample_series:
            # if not validated:
            #     continue
            old_protocol = sample.protocol
            ordered_indices = list(map(lambda x: mix_order_map.get(x.source),sample.protocol))
            argsort = np.argsort(ordered_indices)
            new_protocol = list(map(sample.protocol.__getitem__,argsort))
            time_patched_protocol = []
            for entry in new_protocol:
                if entry.source in custom_stock_settings:
                    for setting,value in custom_stock_settings[entry.source].items():
                        entry.__setattr__(setting,value)
                time_patched_protocol.append(entry)
            sample.protocol = time_patched_protocol
      
   






   



import copy
import warnings
from abc import ABC, abstractmethod
from typing import Optional, Dict, List
import inspect
import json
import importlib

import matplotlib.pyplot as plt
import xarray as xr
from typing_extensions import Self

from AFL.double_agent.PipelineContext import PipelineContext, NoContextException


class PipelineOp(ABC):
    """Abstract base class for data processors. All operations in AFL.double_agent should inherit PipelineOpBase.

    Parameters
    ----------
    name : Optional[str] | List[str]
        The name to use when added to a Pipeline. This name is used when calling Pipeline.search()
    input_variable : Optional[str] | List[str]
        The name of the `xarray.Dataset` data variable to extract from the input dataset
    output_variable : Optional[str] | List[str]
        The name of the variable to be inserted into the `xarray.Dataset` by this `PipelineOp`
    input_prefix : Optional[str] | List[str]
        Prefix for input variables when using pattern matching
    output_prefix : Optional[str] | List[str]
        Prefix for output variables when using pattern matching
    """

    def __init__(
        self,
        name: Optional[str] | List[str] = None,
        input_variable: Optional[str] | List[str] = None,
        output_variable: Optional[str] | List[str] = None,
        input_prefix: Optional[str] | List[str] = None,
        output_prefix: Optional[str] | List[str] = None,
    ):

        if all(
            x is None
            for x in [input_variable, output_variable, input_prefix, output_prefix]
        ):
            warnings.warn(
                "No input/output information set for PipelineOp...this is likely an error",
                stacklevel=2,
            )


        self.name = name or "PipelineOp"
        self.input_variable = input_variable
        self.output_variable = output_variable
        self.input_prefix = input_prefix
        self.output_prefix = output_prefix
        self.output: Dict[str, xr.DataArray] = {}

        # variables to exclude when constructing attrs dict for xarray
        self._banned_from_attrs = ["output", "_banned_from_attrs"]

        ## PipelineContext
        try:
            # try to add this object to current pipeline on context stack
            PipelineContext.get_context().append(self)
        except NoContextException:
            # silently continue for those working outside a context manager
            pass


        ## Gathering Arguments of Most-derived Child Class
        # Retrieve the full stack.
        stack = inspect.stack()
        valid_frames = []
        # Iterate through all frames in the stack.
        for frame_info in stack:
            # Look for __init__ functions where 'self' is our instance.
            if (
                frame_info.function == "__init__"
                and frame_info.frame.f_locals.get("self") is self
            ):
                valid_frames.append(frame_info)
        if valid_frames:
            # Choose the last __init__ call in the inheritance chain.
            final_frame = valid_frames[-1].frame
            args_info = inspect.getargvalues(final_frame)
        else:
            # Fallback: use the immediate caller if nothing was found.
            final_frame = inspect.currentframe().f_back
            args_info = inspect.getargvalues(final_frame)

        # Build _stored_args by checking JSON serializability of each constructor argument.
        stored_args = {}
        for arg in args_info.args:
            if arg != "self":
                value = args_info.locals[arg]
                try:
                    json.dumps(value)
                except (TypeError, OverflowError) as e:
                    raise TypeError(
                        f"Constructor argument '{arg}' with value {value!r} of type {type(value).__name__} "
                        f"is not JSON serializable: {e}"
                    )
                stored_args[arg] = value
        self._stored_args = stored_args

    def __getattribute__(self, name):
        if name == "_stored_args":
            try:
                return object.__getattribute__(self, name)
            except AttributeError:
                return {}
        try:
            stored_args = object.__getattribute__(self, "_stored_args")
        except AttributeError:
            return object.__getattribute__(self, name)
        if name in stored_args:
            return stored_args[name]
        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name == "_stored_args":
            return object.__setattr__(self, name, value)

        try:
            stored_args = object.__getattribute__(self, "_stored_args")
        except AttributeError:
            stored_args = None

        if stored_args is not None and name in stored_args:
            try:
                json.dumps(value)
            except (TypeError, OverflowError) as e:
                raise TypeError(
                    f"New value for attribute '{name}' with value {value!r} of type {type(value).__name__} "
                    f"is not JSON serializable: {e}"
                )
            stored_args[name] = value
        else:
            object.__setattr__(self, name, value)

    def to_json(self):
        """
        Serializes the fully qualified class name and the constructor arguments
        stored in _stored_args into a JSON string.
        """
        cls = self.__class__
        module = cls.__module__
        qualname = cls.__qualname__
        data = {"class": f"{module}.{qualname}", "args": self._stored_args}
        return data

    @classmethod
    def from_json(cls, json_data):
        """
        Deserializes the JSON string back to an object.
        Dynamically imports the module, gets the class, and instantiates it
        using the stored constructor arguments.
        """
        fqcn = json_data["class"]  # e.g. "module.ClassName"
        args = json_data["args"]
        mod_name, class_name = fqcn.rsplit(".", 1)
        module = importlib.import_module(mod_name)
        klass = getattr(module, class_name)
        return klass(**args)

    @abstractmethod
    def calculate(self, dataset: xr.Dataset) -> Self:
        pass

    def __repr__(self) -> str:
        return f"<PipelineOp:{self.name}>"

    def copy(self) -> Self:
        return copy.deepcopy(self)

    def _prefix_output(self, variable_name: str) -> str:
        prefixed_variable = copy.deepcopy(variable_name)
        if self.output_prefix is not None:
            prefixed_variable = f"{self.output_prefix}_{prefixed_variable}"
        return prefixed_variable

    def _get_attrs(self) -> Dict:
        output_dict = copy.deepcopy(self.__dict__)
        for key in self._banned_from_attrs:
            try:
                del output_dict[key]
            except KeyError:
                pass

        # sanitize
        for key in output_dict.keys():
            output_dict[key] = str(output_dict[key])
            # if output_dict[key] is None:
            #     output_dict[key] = 'None'
            # elif type(output_dict[key]) is bool:
            #     output_dict[key] = str(output_dict[key])

        return output_dict

    def _get_variable(self, dataset: xr.Dataset) -> xr.DataArray:
        if self.input_variable is None and self.input_prefix is None:
            raise ValueError(
                (
                    """Can't get variable for {self.name} without input_variable """
                    """or input_prefix specified in constructor """
                )
            )

        if self.input_variable is not None and self.input_prefix is not None:
            raise ValueError(
                (
                    """Both input_variable and input_prefix were specified in constructor. """
                    """Only one should be specified to avoid ambiguous operation"""
                )
            )

        if self.input_variable is not None:
            output = dataset[self.input_variable].copy()

        elif self.input_prefix is not None:
            raise NotImplementedError
        else:
            output = None

        return output

    def add_to_dataset(self, dataset, copy_dataset=True):
        """Adds (xarray) data in output dictionary to provided xarray dataset"""
        if copy_dataset:
            dataset1 = dataset.copy()
        else:
            dataset1 = dataset

        for name, value in self.output.items():
            if isinstance(value, xr.Dataset):
                # add PipelineOp meta variables to attributes
                for data_var in value:
                    value[data_var].attrs.update(self._get_attrs())
                dataset1 = xr.merge([dataset1, value])
            elif isinstance(value, xr.DataArray):
                # add PipelineOp meta variables to attributes
                value.attrs.update(self._get_attrs())
                dataset1[name] = value
            else:
                raise ValueError(
                    (
                        f"""Items in output dictionary of PipelineOp {self.name} must be xr.Dataset or xr.DataArray """
                        f"""Found variable named {name} of type {type(value)}."""
                    )
                )
        return dataset1

    def add_to_tiled(self, tiled_data):
        """Adds data in output dictionary to provided tiled catalogue"""
        raise NotImplementedError
        # This needs to handle/deconstruct xarray types!!
        # for name, dataarray in self.output.items():
        #     tiled_data.add_array(name, value.values)

    def plot(self, sample_dim: str = "sample", **mpl_kwargs) -> plt.Figure:
        """Plots the output of the PipelineOp.

        This method attempts to guess how to plot the data produced by the operation.
        """
        n = len(self.output)
        if n > 0:
            fig, axes = plt.subplots(n, 1, figsize=(6, n * 3))

            if n > 1:
                axes = list(axes.flatten())
            else:
                axes = [axes]

            for i, (name, data) in enumerate(self.output.items()):
                if data.ndim > 1 and (sample_dim in data.dims):
                    data.plot(hue=sample_dim, ax=axes[i], **mpl_kwargs)

                else:
                    data.plot(ax=axes[i], **mpl_kwargs)
                axes[i].set(title=name)
            return fig
        else:
            return plt.figure()



"""
Prefabricated Pipelines module for AFL.double_agent.

This module provides access to prefabricated pipelines that can be used with AFL.double_agent.
It allows loading, listing, and combining multiple pipelines into a single pipeline.
"""

import os
import pathlib
import warnings
import json
from typing import List, Dict, Optional, Union

from AFL.double_agent.Pipeline import Pipeline

# Get the path to the prefab directory
def get_prefab_dir():
    """
    Get the path to the prefabricated pipelines directory.
    
    Returns
    -------
    pathlib.Path
        Path to the prefab directory.
    """
    # The prefab directory is located in AFL/double_agent/prefab
    module_dir = pathlib.Path(__file__).parent
    prefab_dir = module_dir
    
    if not prefab_dir.exists():
        warnings.warn(f"Prefab directory not found at {prefab_dir}")
    
    return prefab_dir

def list_prefabs(display_table: bool = True):
    """
    List all available prefabricated pipelines.
    
    Parameters
    ----------
    display_table : bool, default=True
        Whether to display the results in a formatted table with descriptions.
        If False, returns just the list of prefab names.
    
    Returns
    -------
    list or None
        If display_table is False, returns a list of available prefabricated pipeline names.
        If display_table is True, prints a formatted table and returns None.
    """
    prefab_dir = get_prefab_dir()
    if not prefab_dir.exists():
        warnings.warn(f"Prefab directory not found at {prefab_dir}")
        return []
    
    prefab_files = list(prefab_dir.glob("*.json"))
    
    if not display_table:
        return [f.stem for f in prefab_files]
    
    if not prefab_files:
        print("No prefabricated pipelines found.")
        return None
    
    # Get descriptions from each prefab file
    prefabs_info = []
    max_name_len = 4  # Minimum width for "Name" column
    max_desc_len = 11  # Minimum width for "Description" column
    
    for file_path in prefab_files:
        try:
            with open(file_path, 'r') as f:
                prefab_data = json.load(f)
            
            name = file_path.stem
            description = prefab_data.get('description', 'No description available')
            
            # Track maximum lengths for formatting
            max_name_len = max(max_name_len, len(name))
            max_desc_len = max(max_desc_len, len(description))
            
            prefabs_info.append((name, description))
        except Exception as e:
            warnings.warn(f"Error reading prefab {file_path.name}: {str(e)}")
    
    # Print formatted table
    header = f"| {'Name':<{max_name_len}} | {'Description':<{max_desc_len}} |"
    separator = f"|-{'-'*max_name_len}-|-{'-'*max_desc_len}-|"
    
    print("\nAvailable Prefabricated Pipelines:")
    print(separator)
    print(header)
    print(separator)
    
    for name, description in sorted(prefabs_info):
        print(f"| {name:<{max_name_len}} | {description:<{max_desc_len}} |")
    
    print(separator)
    print(f"Total: {len(prefabs_info)} prefabricated pipeline(s)")
    
    return None

def load_prefab(name: str) -> Pipeline:
    """
    Load a prefabricated pipeline by name.
    
    Parameters
    ----------
    name : str
        Name of the prefabricated pipeline to load.
        
    Returns
    -------
    Pipeline
        The loaded pipeline.
        
    Raises
    ------
    FileNotFoundError
        If the prefabricated pipeline does not exist.
    """
    prefab_dir = get_prefab_dir()
    file_path = prefab_dir / f"{name}.json"
    
    if not file_path.exists():
        raise FileNotFoundError(
            f"Prefabricated pipeline '{name}' not found at {file_path}. "
            f"Prefab directory: {prefab_dir}. "
            f"Available prefabs: {list_prefabs()}"
        )
    
    return Pipeline.read_json(str(file_path))

def combine_prefabs(prefab_names: List[str], new_name: str | None = None) -> Pipeline:
    """
    Combine multiple prefabricated pipelines into a single pipeline.
    
    Parameters
    ----------
    prefab_names : List[str]
        List of prefabricated pipeline names to combine.
    new_name : str | None, default=None
        Name for the combined pipeline. If None, a name will be generated from the component pipelines.
        
    Returns
    -------
    Pipeline
        The combined pipeline.
        
    Raises
    ------
    FileNotFoundError
        If any of the prefabricated pipelines do not exist.
    ValueError
        If no prefabricated pipelines are provided.
    """
    if not prefab_names:
        raise ValueError("No prefabricated pipelines provided for combination.")
    
    if len(prefab_names) == 1:
        return load_prefab(prefab_names[0])
    
    # Generate a combined name if not provided
    if new_name is None:
        new_name = f"Combined_{'_'.join(prefab_names)}"
    
    # Load all pipelines
    pipelines = [load_prefab(name) for name in prefab_names]
    
    # Create a new pipeline with the first pipeline's ops
    combined_pipeline = pipelines[0].copy()
    combined_pipeline.name = new_name
    
    # Add all operations from subsequent pipelines
    for pipeline in pipelines[1:]:
        combined_pipeline.extend(pipeline.ops)
    
    return combined_pipeline

def save_prefab(pipeline: Pipeline, name: str | None = None, overwrite: bool = False, description: str | None = None) -> str:
    """
    Save a pipeline as a prefabricated pipeline.
    
    Parameters
    ----------
    pipeline : Pipeline
        The pipeline to save.
    name : str | None, default=None
        Name to save the pipeline as. If None, uses the pipeline's name.
    overwrite : bool, default=False
        Whether to overwrite an existing prefabricated pipeline with the same name.
    description : str | None, default=None
        A descriptive text about the pipeline's purpose and functionality.
        If None and pipeline has a description attribute, that will be used.
        
    Returns
    -------
    str
        The path to the saved prefabricated pipeline file.
        
    Raises
    ------
    FileExistsError
        If a prefabricated pipeline with the given name already exists and overwrite=False.
    """
    prefab_dir = get_prefab_dir()
    
    # Use the pipeline's name if no name is provided
    if name is None:
        name = pipeline.name
    
    # Use the pipeline's description if available and none provided
    if description is None and hasattr(pipeline, 'description'):
        description = pipeline.description
    
    # Ensure valid filename
    name = name.replace(" ", "_")
    file_path = prefab_dir / f"{name}.json"
    
    # Check if file exists and overwrite is not allowed
    if file_path.exists() and not overwrite:
        raise FileExistsError(
            f"Prefabricated pipeline '{name}' already exists at {file_path}. "
            f"Use overwrite=True to overwrite."
        )
    
    # Save the pipeline with description
    pipeline.write_json(str(file_path), overwrite=True, description=description)
    
    return str(file_path)

# Define any example prefabs that should be available by default
def example_prefab1():
    """
    Load an example prefabricated pipeline.
    
    Returns
    -------
    Pipeline
        An example prefabricated pipeline.
    """
    try:
        return load_prefab("example_prefab")
    except FileNotFoundError:
        warnings.warn("Example prefab not found. You may need to create example prefabs.")
        return Pipeline(name="empty_example")

# Export public functions
__all__ = [
    "get_prefab_dir", 
    "list_prefabs", 
    "load_prefab", 
    "combine_prefabs", 
    "save_prefab", 
    "example_prefab1"
] 


import ipywidgets
import numpy as np
import plotly.graph_objects as go
from typing import Dict, List, Any, Optional, Tuple, Set, Union
import copy
import uuid
import json
import sasmodels.data
import sasmodels.direct_model
import sasmodels.core
import ipywidgets as widgets
from IPython.display import display, JSON
import re
import xarray as xr

# Make sure plotly works well in Jupyter
#from plotly.offline import init_notebook_mode
#init_notebook_mode(connected=True)

class AutoSASWidget:
    """Interactive widget for building SAS model inputs for AutoSAS.
    
    This widget allows users to create, configure and visualize SAS models
    through an interactive interface. Users can add multiple models as tabs,
    adjust their parameters, and see real-time visualization of the models.
    
    The interface follows an MVC pattern with separate model and view classes.
    """
    
    def __init__(
        self,
        q_range: Optional[np.ndarray] = None,
        default_models: Optional[List[Dict[str, Any]]] = None,
        data: Optional[xr.Dataset] = None,
        model_inputs: Optional[List[Dict[str, Any]]] = None,
        custom_models = None,
    ):
        """Initialize the AutoSASWidget.
        
        Parameters
        ----------
        q_range : Optional[np.ndarray]
            Array of q values to use for model visualization.
            If None, a default range will be created.
        default_models : Optional[List[Dict[str, Any]]]
            List of model configurations to initialize with.
        data : Optional[xr.Dataset]
            Dataset containing experimental scattering data.
        model_inputs : Optional[List[Dict[str, Any]]]
            List of model configurations to initialize with. Takes precedence over default_models.
        """
        # Set up default q_range if not provided
        if q_range is None:
            self.q_range = np.logspace(-3, 0, 100)
        else:
            self.q_range = q_range
            
        # Create the model and view
        self.model = AutoSASWidget_Model(q_range=self.q_range, default_models=default_models,custom_models=custom_models)
        self.view = AutoSASWidget_View(available_models=self.model.available_sasmodels, data=data)
        
        # Connect model and view
        self._setup_callbacks()
        
        # Load model inputs if provided
        if model_inputs:
            self.load_model_inputs(model_inputs)
    
    def _setup_callbacks(self):
        """Connect UI events to model methods."""
        # Add model tab button callback
        self.view.button["add_model"].on_click(self.add_model_callback)
        
        # Remove model tab button callback
        self.view.button["remove_model"].on_click(self.remove_model_callback)
        
        # Export button callback
        self.view.button["export"].on_click(self.export_callback)
    
    def add_model_callback(self, b):
        """Handle add model button click."""
        # Get the model name from dropdown
        model_name = self.view.dropdown["model_type"].value
        
        # Add model to the model component
        model_id = self.model.add_model(model_name)
        
        # Add a new tab to the view
        self.view.add_model_tab(model_id, model_name)
        
        # Set up parameter controls for the new model
        self._setup_parameter_controls(model_id)
        
        # Update the plot
        self.update_plot(model_id)

        self.view.figures[model_id].update_layout(height=549)
        self.view.figures[model_id].update_layout(height=550)

        # Set the new tab as active
        self.view.tabs.selected_index = len(self.view.tab_ids) - 1
    
    def remove_model_callback(self, b):
        """Handle remove model button click."""
        # Get currently selected tab
        current_tab = self.view.tabs.selected_index
        if current_tab >= 0:
            # Get model ID
            model_id = self.view.tab_ids[current_tab]
            
            # Remove from model
            self.model.remove_model(model_id)
            
            # Remove tab from view
            self.view.remove_model_tab(current_tab)
            
            # Update titles for remaining tabs
            for i, tab_id in enumerate(self.view.tab_ids):
                # Get the base model name (remove the _N suffix)
                base_name = self.model.models[tab_id]["sasmodel"]
                # Update the tab title with new index
                self.view.tabs.set_title(i, f"{base_name}_{i + 1}")
                # Update the model name in the model data
                self.model.models[tab_id]["name"] = f"{base_name}_{i + 1}"
    
    def _setup_parameter_controls(self, model_id):
        """Set up parameter controls for a model."""
        # Get all parameters for this model
        params = self.model.get_model_parameters(model_id)
        
        # Set up q_min and q_max controls
        q_min_control = self.view.parameter_controls[model_id]["q_min"]
        q_max_control = self.view.parameter_controls[model_id]["q_max"]
        
        # Get current q range from model
        model_data = self.model.models[model_id]
        q_min_control.value = model_data["q_min"]
        q_max_control.value = model_data["q_max"]
        
        # Set up callbacks for q range controls
        q_min_control.observe(
            lambda change, mid=model_id: 
                self.q_range_change_callback(change, mid, "q_min"),
            names="value"
        )
        
        q_max_control.observe(
            lambda change, mid=model_id: 
                self.q_range_change_callback(change, mid, "q_max"),
            names="value"
        )
        
        # Add controls for each parameter
        for param_name, param_info in params.items():
            # Determine if this parameter should be included in AutoSAS
            use_autosas = param_info.get("autosas", False)
            
            # Determine if this parameter should use bounds
            use_bounds = param_info.get("use_bounds", False)
            
            # Add the control
            self.view.add_parameter_control(
                model_id, 
                param_name, 
                param_info["value"],
                param_info.get("bounds"),
                use_autosas=use_autosas,
                use_bounds=use_bounds
            )
            
            # Get the control and set up callback
            control = self.view.get_parameter_control(model_id, param_name)
            control.observe(
                lambda change, mid=model_id, pname=param_name: 
                    self.parameter_change_callback(change, mid, pname),
                names="value"
            )
            
            # Get the AutoSAS checkbox and set up callback
            autosas_checkbox = self.view.get_parameter_checkbox(model_id, param_name)
            if autosas_checkbox:
                autosas_checkbox.observe(
                    lambda change, mid=model_id, pname=param_name: 
                        self.parameter_checkbox_callback(change, mid, pname),
                    names="value"
                )
            
            # Get the bounds checkbox and set up callback
            bounds_checkbox = self.view.get_bounds_checkbox(model_id, param_name)
            if bounds_checkbox:
                bounds_checkbox.observe(
                    lambda change, mid=model_id, pname=param_name: 
                        self.parameter_checkbox_callback(change, mid, pname),
                    names="value"
                )
    
    def parameter_change_callback(self, change, model_id, param_name):
        """Handle parameter value change."""
        # Update the model parameter
        self.model.update_parameter(model_id, param_name, change["new"])
        
        # Update the plot
        self.update_plot(model_id)
    
    def parameter_checkbox_callback(self, change, model_id, param_name):
        """Handle parameter checkbox change (fixed/variable)."""
        # Determine which checkbox was changed (AutoSAS or Bounds)
        checkbox_type = change.get('owner').description.lower()
        
        if checkbox_type == 'autosas':
            # Update model parameter to reflect AutoSAS mode (include in model_inputs)
            self.model.models[model_id]["params"][param_name]["autosas"] = change["new"]
        elif checkbox_type == 'bounds':
            # Update model parameter to reflect Bounds mode (show slider and include bounds)
            self.model.models[model_id]["params"][param_name]["use_bounds"] = change["new"]
            
            # Get current controls
            control = self.view.get_parameter_control(model_id, param_name)
            
            # Get parameter bounds
            bounds = self.model.get_parameter_bounds(model_id, param_name)
            
            # If bounds checkbox is checked, replace with slider
            if change["new"]:
                self.view.replace_with_slider(model_id, param_name, control.value, bounds)
            # If bounds checkbox is unchecked, replace with text input
            else:
                self.view.replace_with_text(model_id, param_name, control.value)
        
        # Update the plot after any checkbox change
        self.update_plot(model_id)
    
    def update_plot(self, model_id):
        """Update the plot for a model."""
        # Calculate model intensity
        q, intensity = self.model.calculate_model_intensity(model_id)
        
        # Update the plot
        self.view.update_plot(model_id, q, intensity)
    
    def get_model_inputs(self) -> List[Dict[str, Any]]:
        """Get the model inputs in the format required by AutoSAS.
        
        Returns
        -------
        List[Dict[str, Any]]
            List of model configurations
        """
        result = []
        
        for model_id, model in self.model.models.items():
            # Convert our internal format to AutoSAS format
            fit_params = {}
            # Group parameters by their base name (for polydispersity)
            param_groups = {}
            
            for name, param_info in model["params"].items():
                # Only include parameter if AutoSAS is enabled
                if param_info.get("autosas", False):
                    # Check if this is a polydispersity parameter
                    if name.endswith(('_pd', '_pd_type', '_pd_n', '_pd_nsigma')):
                        # Get the base parameter name
                        base_name = name.rsplit('_pd', 1)[0]
                        if base_name not in param_groups:
                            param_groups[base_name] = {}
                        
                        # Store the polydispersity parameter
                        if name.endswith('_pd'):
                            param_groups[base_name]['pd'] = param_info["value"]
                        elif name.endswith('_pd_type'):
                            param_groups[base_name]['pd_type'] = param_info["value"]
                        elif name.endswith('_pd_n'):
                            param_groups[base_name]['pd_n'] = param_info["value"]
                        elif name.endswith('_pd_nsigma'):
                            param_groups[base_name]['pd_nsigma'] = param_info["value"]
                    else:
                        # Regular parameter
                        param_dict = {"value": param_info["value"]}
                        
                        # Only include bounds if bounds mode is enabled and bounds exist
                        if (param_info.get("use_bounds", False) and 
                            param_info.get("bounds") is not None):
                            param_dict["bounds"] = param_info["bounds"]
                        
                        fit_params[name] = param_dict
            
            # Add polydispersity parameters to their base parameters
            for base_name, pd_params in param_groups.items():
                if base_name in fit_params:
                    # Only add polydispersity if the base parameter is included
                    for pd_key, pd_value in pd_params.items():
                        fit_params[base_name][pd_key] = pd_value
            
            # Create the model config
            model_config = {
                "name": model["name"],
                "sasmodel": model["sasmodel"],
                "q_min": model["q_min"],
                "q_max": model["q_max"],
                "fit_params": fit_params
            }
            
            result.append(model_config)
        
        return result
    
    def run(self):
        """Display the widget interface."""
        display(self.view.main_container)
    
    def export_callback(self, b):
        """Handle export button click."""
        # Get model inputs and store in widget attribute
        self.model_inputs = self.get_model_inputs()
        
        # Show notification
        self.view.notification.value = 'model inputs written to widget.model_inputs'
        self.view.notification.layout.display = 'block'
        
        # Schedule notification to disappear after 5 seconds
        import asyncio
        from IPython.display import display
        
        async def hide_notification():
            await asyncio.sleep(5)
            self.view.notification.layout.display = 'none'
        
        asyncio.create_task(hide_notification())
    
    def load_model_inputs(self, model_inputs: List[Dict[str, Any]]) -> None:
        """Load model inputs from an existing configuration.
        
        This allows users to load existing AutoSAS model configurations
        into the widget for further editing and visualization.
        
        Parameters
        ----------
        model_inputs : List[Dict[str, Any]]
            List of model configurations in AutoSAS format
        """
        # Clear existing models first
        # Get all tab IDs
        tab_ids = self.view.tab_ids.copy()
        
        # Remove all tabs
        for model_id in tab_ids:
            self.model.remove_model(model_id)
            
        # Clear view tabs
        self.view.clear_tabs()
        
        # Add models from configuration
        for model_config in model_inputs:
            # Add to model
            model_id = self.model.add_model_from_config(model_config)
            
            # Add to view
            self.view.add_model_tab(model_id, model_config["name"])
            
            # Set up parameter controls
            self._setup_parameter_controls(model_id)
            
            # Update plot
            self.update_plot(model_id)
    
    def q_range_change_callback(self, change, model_id, range_type):
        """Handle q range change."""
        # Update the model q range
        if range_type == "q_min":
            self.model.update_q_range(model_id, q_min=change["new"])
        else:
            self.model.update_q_range(model_id, q_max=change["new"])
        
        # Update the plot
        self.update_plot(model_id)


class AutoSASWidget_Model:
    """Model component for AutoSASWidget.
    
    Handles the data and business logic for the SAS models.
    """
    
    def __init__(
        self,
        q_range: np.ndarray,
        default_models: Optional[List[Dict[str, Any]]] = None,
        custom_models = None
    ):
        """Initialize the model component.
        
        Parameters
        ----------
        q_range : np.ndarray
            Array of q values to use for model visualization
        default_models : Optional[List[Dict[str, Any]]]
            List of model configurations to initialize with
        """
        self.q_range = q_range
        self.models = {}

        self.custom_models = custom_models or []
        self.available_sasmodels = custom_models + self._get_available_sasmodels() 
        
        # Initialize with default models if provided
        if default_models:
            for model_config in default_models:
                self.add_model_from_config(model_config)
    
    def _get_available_sasmodels(self) -> List[str]:
        """Get list of available SAS models from sasmodels.
        
        Returns
        -------
        List[str]
            List of available model names
        """
        try:
            return sorted(sasmodels.core.list_models())
        except Exception as e:
            print(f"Error getting model list: {e}")
            # Fallback to a known list of common models
            return [
                "sphere", "cylinder", "ellipsoid", "core_shell_sphere",
                "core_shell_cylinder", "power_law", "fractal", "pearl_necklace",
                "guinier", "guinier_porod", "unified_power_rg", "dab", 
                "broad_peak", "two_lorentzian", "teubner_strey"
            ]
    
    def add_model(self, model_name: str) -> str:
        """Add a new SAS model.
        
        Parameters
        ----------
        model_name : str
            Name of the SAS model
            
        Returns
        -------
        str
            Unique ID for the created model
        """
        # Generate a unique ID for this model
        model_id = str(uuid.uuid4())
        
        # Create model kernel
        kernel = sasmodels.core.load_model(model_name)
        
        # Get polydispersity parameters
        pd_params = kernel.info.parameters.pd_1d if hasattr(kernel.info.parameters, 'pd_1d') else []
        
        # Get default parameters
        params = {}
        for name, value in kernel.info.parameters.defaults.items():
            # Determine reasonable bounds based on the parameter
            # This is a simple heuristic and might need adjustment
            if value > 0:
                bounds = (value / 10, value * 10)
            elif value < 0:
                bounds = (value * 10, value / 10)
            else:
                bounds = (-1, 1)
            
            params[name] = {
                "value": value,
                "bounds": bounds,
                "fixed": False
            }
            
            # Add polydispersity parameters if this parameter supports it
            if name in pd_params:
                # Add width parameter
                params[f"{name}_pd"] = {
                    "value": 0.0,
                    "bounds": (0.0, 1.0),
                    "fixed": False
                }
                
                # Add distribution type parameter
                params[f"{name}_pd_type"] = {
                    "value": "gaussian",
                    "options": ["gaussian", "rectangular", "schulz"],
                    "fixed": False
                }
                
                # Add number of points parameter
                params[f"{name}_pd_n"] = {
                    "value": 35,
                    "bounds": (3, 200),
                    "fixed": False
                }
                
                # Add number of sigmas parameter
                params[f"{name}_pd_nsigma"] = {
                    "value": 3.0,
                    "bounds": (1.0, 10.0),
                    "fixed": False
                }
        
        # Store the model
        self.models[model_id] = {
            "name": f"{model_name}_{len(self.models) + 1}",
            "sasmodel": model_name,
            "kernel": kernel,
            "params": params,
            "q_min": min(self.q_range),
            "q_max": max(self.q_range)
        }
        
        return model_id
    
    def add_model_from_config(self, model_config: Dict[str, Any]) -> str:
        """Add a model from an existing configuration.
        
        Parameters
        ----------
        model_config : Dict[str, Any]
            Model configuration dictionary
            
        Returns
        -------
        str
            Unique ID for the created model
        """
        # Generate a unique ID for this model
        model_id = str(uuid.uuid4())
        
        # Create model kernel
        kernel = sasmodels.core.load_model(model_config["sasmodel"])
        
        # Get polydispersity parameters
        pd_params = kernel.info.parameters.pd_1d if hasattr(kernel.info.parameters, 'pd_1d') else []
        
        # Start with all default parameters
        params = {}
        for name, value in kernel.info.parameters.defaults.items():
            # Determine reasonable bounds based on the parameter
            # This is a simple heuristic and might need adjustment
            if value > 0:
                bounds = (value / 10, value * 10)
            elif value < 0:
                bounds = (value * 10, value / 10)
            else:
                bounds = (-1, 1)
            
            params[name] = {
                "value": value,
                "bounds": bounds,
                "fixed": True,
                "autosas": False,
                "use_bounds": False
            }
            
            # Add polydispersity parameters if this parameter supports it
            if name in pd_params:
                # Add width parameter
                params[f"{name}_pd"] = {
                    "value": 0.0,
                    "bounds": (0.0, 1.0),
                    "fixed": True,
                    "autosas": False,
                    "use_bounds": False
                }
                
                # Add distribution type parameter
                params[f"{name}_pd_type"] = {
                    "value": "gaussian",
                    "options": ["gaussian", "rectangular", "schulz"],
                    "fixed": True,
                    "autosas": False,
                    "use_bounds": False
                }
                
                # Add number of points parameter
                params[f"{name}_pd_n"] = {
                    "value": 35,
                    "bounds": (3, 200),
                    "fixed": True,
                    "autosas": False,
                    "use_bounds": False
                }
                
                # Add number of sigmas parameter
                params[f"{name}_pd_nsigma"] = {
                    "value": 3.0,
                    "bounds": (1.0, 10.0),
                    "fixed": True,
                    "autosas": False,
                    "use_bounds": False
                }
        
        # Now update with the provided parameters from the configuration
        for name, param_info in model_config["fit_params"].items():
            # Update base parameter
            if name in params:
                params[name].update({
                    "value": param_info["value"],
                    "bounds": param_info.get("bounds", params[name]["bounds"]),
                    "fixed": "bounds" not in param_info,
                    "autosas": True,  # If it's in fit_params, it should be included in AutoSAS
                    "use_bounds": "bounds" in param_info
                })
            
            # Handle polydispersity parameters if present
            if any(pd_key in param_info for pd_key in ['pd', 'pd_type', 'pd_n', 'pd_nsigma']):
                if 'pd' in param_info and f"{name}_pd" in params:
                    params[f"{name}_pd"].update({
                        "value": param_info["pd"],
                        "autosas": True
                    })
                if 'pd_type' in param_info and f"{name}_pd_type" in params:
                    params[f"{name}_pd_type"].update({
                        "value": param_info["pd_type"],
                        "autosas": True
                    })
                if 'pd_n' in param_info and f"{name}_pd_n" in params:
                    params[f"{name}_pd_n"].update({
                        "value": param_info["pd_n"],
                        "autosas": True
                    })
                if 'pd_nsigma' in param_info and f"{name}_pd_nsigma" in params:
                    params[f"{name}_pd_nsigma"].update({
                        "value": param_info["pd_nsigma"],
                        "autosas": True
                    })
        
        # Store the model
        self.models[model_id] = {
            "name": model_config["name"],
            "sasmodel": model_config["sasmodel"],
            "kernel": kernel,
            "params": params,
            "q_min": model_config.get("q_min", min(self.q_range)),
            "q_max": model_config.get("q_max", max(self.q_range))
        }
        
        return model_id
    
    def remove_model(self, model_id: str) -> None:
        """Remove a model.
        
        Parameters
        ----------
        model_id : str
            ID of the model to remove
        """
        if model_id in self.models:
            del self.models[model_id]
    
    def get_model_parameters(self, model_id: str) -> Dict[str, Dict[str, Any]]:
        """Get parameters for a model.
        
        Parameters
        ----------
        model_id : str
            ID of the model
            
        Returns
        -------
        Dict[str, Dict[str, Any]]
            Dictionary of parameter information
        """
        if model_id in self.models:
            return self.models[model_id]["params"]
        return {}
    
    def get_parameter_bounds(self, model_id: str, param_name: str) -> Tuple[float, float]:
        """Get bounds for a parameter.
        
        Parameters
        ----------
        model_id : str
            ID of the model
        param_name : str
            Name of the parameter
            
        Returns
        -------
        Tuple[float, float]
            Lower and upper bounds for the parameter
        """
        if model_id in self.models and param_name in self.models[model_id]["params"]:
            bounds = self.models[model_id]["params"][param_name].get("bounds")
            if bounds:
                return bounds
            
            # Default bounds if none specified
            value = self.models[model_id]["params"][param_name]["value"]
            if value > 0:
                return (value / 10, value * 10)
            elif value < 0:
                return (value * 10, value / 10)
            else:
                return (-1, 1)
        
        return (-1, 1)  # Default fallback
    
    def update_parameter(self, model_id: str, param_name: str, value: float) -> None:
        """Update a parameter value.
        
        Parameters
        ----------
        model_id : str
            ID of the model
        param_name : str
            Name of the parameter
        value : float
            New value for the parameter
        """
        if model_id in self.models and param_name in self.models[model_id]["params"]:
            self.models[model_id]["params"][param_name]["value"] = value
    
    def calculate_model_intensity(self, model_id: str) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate the model intensity.
        
        Parameters
        ----------
        model_id : str
            ID of the model
            
        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            q values and corresponding intensities
        """
        if model_id not in self.models:
            return self.q_range, np.zeros_like(self.q_range)
        
        model = self.models[model_id]
        kernel = model["kernel"]
        
        # Create parameter dictionary for calculation
        params = {}
        for name, param_info in model["params"].items():
            name = str(name).replace("<", "").replace(">", "")
            params[str(name)] = param_info["value"]
        
        # Apply q range filtering
        q_mask = (self.q_range >= model["q_min"]) & (self.q_range <= model["q_max"])
        q_values = self.q_range[q_mask]
        
        try:
            # Calculate the model
            # Create empty data with resolution
            data = sasmodels.data.empty_data1D(q_values)
            # Create direct model calculator
            calculator = sasmodels.direct_model.DirectModel(data, kernel)
            # Calculate intensity
            intensity = calculator(**params)
            
            return q_values, intensity
        except Exception as e:
            print(f"Error calculating model: {e}")
            # Return empty arrays on error
            return np.array([]), np.array([])
    
    def get_model_inputs(self) -> List[Dict[str, Any]]:
        """Get the model inputs in the format required by AutoSAS.
        
        Returns
        -------
        List[Dict[str, Any]]
            List of model configurations
        """
        result = []
        
        for model_id, model in self.models.items():
            # Convert our internal format to AutoSAS format
            fit_params = {}
            # Group parameters by their base name (for polydispersity)
            param_groups = {}
            
            for name, param_info in model["params"].items():
                # Only include parameter if AutoSAS is enabled
                if param_info.get("autosas", False):
                    # Check if this is a polydispersity parameter
                    if name.endswith(('_pd', '_pd_type', '_pd_n', '_pd_nsigma')):
                        # Get the base parameter name
                        base_name = name.rsplit('_pd', 1)[0]
                        if base_name not in param_groups:
                            param_groups[base_name] = {}
                        
                        # Store the polydispersity parameter
                        if name.endswith('_pd'):
                            param_groups[base_name]['pd'] = param_info["value"]
                        elif name.endswith('_pd_type'):
                            param_groups[base_name]['pd_type'] = param_info["value"]
                        elif name.endswith('_pd_n'):
                            param_groups[base_name]['pd_n'] = param_info["value"]
                        elif name.endswith('_pd_nsigma'):
                            param_groups[base_name]['pd_nsigma'] = param_info["value"]
                    else:
                        # Regular parameter
                        param_dict = {"value": param_info["value"]}
                        
                        # Only include bounds if bounds mode is enabled and bounds exist
                        if (param_info.get("use_bounds", False) and 
                            param_info.get("bounds") is not None):
                            param_dict["bounds"] = param_info["bounds"]
                        
                        fit_params[name] = param_dict
            
            # Add polydispersity parameters to their base parameters
            for base_name, pd_params in param_groups.items():
                if base_name in fit_params:
                    # Only add polydispersity if the base parameter is included
                    for pd_key, pd_value in pd_params.items():
                        fit_params[base_name][pd_key] = pd_value
            
            # Create the model config
            model_config = {
                "name": model["name"],
                "sasmodel": model["sasmodel"],
                "q_min": model["q_min"],
                "q_max": model["q_max"],
                "fit_params": fit_params
            }
            
            result.append(model_config)
        
        return result
    
    def update_q_range(self, model_id: str, q_min: float = None, q_max: float = None) -> None:
        """Update the q range for a model.
        
        Parameters
        ----------
        model_id : str
            ID of the model
        q_min : float, optional
            Minimum q value
        q_max : float, optional
            Maximum q value
        """
        if model_id in self.models:
            if q_min is not None:
                self.models[model_id]["q_min"] = q_min
            if q_max is not None:
                self.models[model_id]["q_max"] = q_max


class AutoSASWidget_View:
    """View component for AutoSASWidget.
    
    Handles the UI elements and their layout.
    """
    
    def __init__(
        self,
        available_models=None,
        data: Optional[xr.Dataset] = None
    ):
        """Initialize the view component.
        
        Parameters
        ----------
        available_models : List[str], optional
            List of available SAS model names
        data : Optional[xr.Dataset]
            Dataset containing experimental scattering data.
        """
        # Containers for UI elements
        self.tabs = widgets.Tab()
        self.dropdown = {}
        self.button = {}
        self.text_input = {}
        self.checkbox = {}
        self.figures = {}
        
        # Available models
        self.available_models = available_models or [
            "sphere", "cylinder", "ellipsoid", "core_shell_sphere",
            "core_shell_cylinder", "power_law"
        ]
        
        # Track the tab IDs
        self.tab_ids = []
        
        # Track parameter controls for each model
        self.parameter_controls = {}
        self.parameter_checkboxes = {}
        
        # Store the dataset
        self.data = data
        
        # Initialize the UI
        self._init_ui()
    
    def _init_ui(self):
        """Initialize the UI elements."""
        # Create title
        title = widgets.HTML(
            "<div style='background-color: #f8f9fa; padding: 10px; border-bottom: 2px solid #dee2e6;'>"
            "<h2 style='margin: 0;'>AutoSAS Model Builder</h2>"
            "</div>"
        )
        
        # Create data selection controls if data is provided
        data_controls = None
        if self.data is not None:
            # Get data variables that could be intensity data (1D or 2D arrays)
            data_vars = [var for var, da in self.data.data_vars.items() 
                        if len(da.dims) in [1, 2]]
            data_vars = [var for var, da in self.data.data_vars.items() 
                        if len(da.dims) in [1, 2]]
            # Create data variable dropdown
            self.dropdown["data_var"] = widgets.Dropdown(
                options=data_vars,
                description="Data:",
                value=data_vars[0] if data_vars else None,
                layout=widgets.Layout(width='200px')
            )
            
            # Create q dimension dropdown
            self.dropdown["q_dim"] = widgets.Dropdown(
                options=[],
                description="q dim:",
                layout=widgets.Layout(width='200px')
            )
            
            # Create data index slider
            self.text_input["data_index"] = widgets.IntSlider(
                value=0,
                min=0,
                max=1,
                step=0.1,
                description="Index:",
                #continuous_update=True,  # Add this to update while sliding
                #readout_format='.0f',    # Show as integer
                layout=widgets.Layout(width='300px')
            )
            
            # Update controls when data variable changes
            def update_data_controls(change):
                da = self.data[change["new"]]
                
                # Update q dimension options
                q_dims = [dim for dim in da.dims if 'q' in dim.lower()]
                self.dropdown["q_dim"].options = q_dims
                if q_dims:
                    self.dropdown["q_dim"].value = q_dims[0]
                
                # Update index slider based on non-q dimension
                non_q_dims = [dim for dim in da.dims if 'q' not in dim.lower()]
                if non_q_dims:
                    index_dim = non_q_dims[0]
                    index_size = da.sizes[index_dim]
                    self.text_input["data_index"].max = index_size - 1
                    self.text_input["data_index"].step = 1
                    self.text_input["data_index"].value = 0  # Reset to first index
                
                # Update plots
                for model_id in self.figures:
                    self.update_plot(model_id, [], [])
            
            # Update plot when any control changes
            def update_data_display(*args):
                for model_id in self.figures:
                    self.update_plot(model_id, [], [])
            
            # Connect all data control observers
            self.dropdown["data_var"].observe(update_data_controls, names="value")
            self.dropdown["q_dim"].observe(update_data_display, names="value")
            self.text_input["data_index"].observe(update_data_display, names="value")
            self.button["refresh_data"] = widgets.Button(
                            description="Refresh Data",
                            button_style="info",
                            icon='refresh',
                            layout=widgets.Layout(width='120px', margin='5px 0px 0px 10px')
                        )
            
            # Initialize controls with first data variable
            if data_vars:
                update_data_controls({"new": data_vars[0]})
            
            # Create data controls container
            data_controls = widgets.VBox([
                widgets.HTML("<b>Select Data:</b>"),
                widgets.HBox([
                    widgets.VBox([
                        self.dropdown["data_var"], 
                        self.dropdown["q_dim"],
                        self.text_input["data_index"],
                    ]),
                ], layout=widgets.Layout(margin='5px 0px 15px 10px'))
            ], layout=widgets.Layout(margin='10px 0px', padding='10px',
                                   border='1px solid #eee'))

            # Add observer for refresh button
            def refresh_data(b):
                for model_id in self.figures:
                    self.update_plot(model_id, [], [])
            
            self.button["refresh_data"].on_click(refresh_data)
        
        # Create the model selection dropdown with search
        self.dropdown["model_type"] = widgets.Select(
            options=self.available_models,
            value=self.available_models[0] if self.available_models else "sphere",
            description="Model Type:",
            layout=widgets.Layout(width='300px', height='100px'),
            search=True,  # Enable fuzzy search
            ensure_option=True
        )
        
        # Create add/remove model buttons
        self.button["add_model"] = widgets.Button(
            description="Add Model",
            button_style="success",
            icon='plus',
            layout=widgets.Layout(width='120px')
        )
        
        self.button["remove_model"] = widgets.Button(
            description="Remove Model",
            button_style="danger",
            icon='minus',
            layout=widgets.Layout(width='140px')
        )
        
        # Create export button
        self.button["export"] = widgets.Button(
            description="Export Model Inputs",
            button_style="info",
            icon='download',
            layout=widgets.Layout(width='180px')
        )
        
        # Create notification widget (initially hidden)
        self.notification = widgets.HTML(
            value='',
            layout=widgets.Layout(display='none', margin='5px 0px 0px 0px', 
                                color='green', font_style='italic')
        )
        
        # Create model selection container
        model_selection = widgets.VBox([
            widgets.HTML("<b>Select Model:</b>"),
            self.dropdown["model_type"]
        ], layout=widgets.Layout(margin='10px 0px', padding='10px',
                               border='1px solid #eee'))
        
        # Create buttons container
        buttons_container = widgets.VBox([
            self.button["add_model"],
            self.button["remove_model"],
            widgets.VBox([
                self.button["export"],
                self.notification
            ])
        ], layout=widgets.Layout(margin='20px 0px 0px 20px'))
        
        # Create the top control panel with better spacing
        controls = [model_selection, buttons_container]
        if data_controls is not None:
            controls.insert(0, data_controls)
        top_controls = widgets.HBox(
            controls,
            layout=widgets.Layout(margin='10px 0px 20px 0px', justify_content='flex-start')
        )
        
        # Create the tabs container for models with styling
        self.tabs = widgets.Tab(layout=widgets.Layout(width='100%', min_height='600px'))
        
        # Create the main container
        self.main_container = widgets.VBox([
            title,
            top_controls,
            self.tabs
        ], layout=widgets.Layout(width='1050px', padding='20px'))
    
    def add_model_tab(self, model_id: str, model_name: str) -> None:
        """Add a new tab for a model."""
        # Create plot figure with explicit notebook initialization
        fig = go.FigureWidget(layout={
        })
        fig.update_xaxes(type="log",title="q (Å⁻¹)")
        fig.update_yaxes(type="log",title="Intensity (cm⁻¹)")
        fig.update_layout({
            "legend":dict( yanchor="top", y=0.99, xanchor="right", x=0.01), 
            "margin": dict(l=20, r=0, t=40, b=20),
            "title": f"{model_name} Model",
            "template": "plotly_white",
            "showlegend": False,
            "height": 500,  # Increased height
            "width": 500,   # Increased width
        })
        
        # Add initial trace with better styling
        fig.add_scatter(
            x=[], 
            y=[], 
            mode="lines", 
            line=dict(color="royalblue", width=3),
            name="Model",
            hovertemplate="q: %{x:.3e}<br>I: %{y:.3e}<extra></extra>"
        )
        
        # Add trace for experimental data if available
        if self.data is not None:
            fig.add_scatter(
                x=[], y=[], mode="markers", 
                marker=dict(color="black", size=6),
                name="Data",
                hovertemplate="q: %{x:.3e}<br>I: %{y:.3e}<extra></extra>")
        
        # Store the figure
        self.figures[model_id] = fig
        
        # Create title for parameters section with better styling
        params_title = widgets.HTML(
            "<div style='background-color: #f8f9fa; padding: 10px; border-bottom: 2px solid #dee2e6;'>"
            "<h3 style='margin: 0;'>Model Parameters</h3>"
            "</div>"
        )
        
        # Create filter checkboxes
        filter_title = widgets.HTML("<b>Show Parameters:</b>")
        polydispersity_checkbox = widgets.Checkbox(
            value=False,
            description="Polydispersity",
            indent=False,
            layout=widgets.Layout(width='150px', margin='5px 0px 0px 10px')
        )
        magnetic_checkbox = widgets.Checkbox(
            value=False,
            description="Magnetic",
            indent=False,
            layout=widgets.Layout(width='150px', margin='5px 0px 0px 10px')
        )
        
        # Create filter box
        filter_box = widgets.VBox([
            filter_title,
            widgets.HBox([polydispersity_checkbox, magnetic_checkbox],
                layout=widgets.Layout(margin='5px 0px 15px 10px'))
        ],
            layout=widgets.Layout(
                margin='10px 0px',
                border='1px solid #eee',
                padding='0px 0px 10px 0px',
                min_height='30px',
                flex='none'
            )
        )
        
        # Create q range controls with better organization
        q_range_title = widgets.HTML("<b>q range:</b>")
        q_min_input = widgets.FloatText(
            value=0.001,
            description="q min:",
            style={'description_width': 'initial'},
            layout=widgets.Layout(width='180px')
        )
        
        q_max_input = widgets.FloatText(
            value=1.0,
            description="q max:",
            style={'description_width': 'initial'},
            layout=widgets.Layout(width='180px')
        )
        
        q_range_box = widgets.VBox([
            q_range_title,
            widgets.HBox([q_min_input, q_max_input], 
                layout=widgets.Layout(margin='5px 0px 15px 10px')),
        ],
            layout=widgets.Layout(
                margin='10px 0px',
                border='1px solid #eee',
                padding='0px 0px 10px 0px',
                min_height='60px',
                flex='none'
            )
        )
        
        # Create parameters container with better scroll and styling
        params_container = widgets.VBox(
            [params_title, filter_box, q_range_box],
            layout=widgets.Layout(
                width='400px',          # Increased width
                height='550px',         # Match plot height
                overflow_y='scroll',    # Ensure vertical scrolling
                border='1px solid #dee2e6',
                box_shadow='1px 1px 3px rgba(0,0,0,0.1)',
                margin='0px 0px 0px 20px'  # Add margin to separate from plot
            )
        )
        
        # Initialize parameter controls containers
        self.parameter_controls[model_id] = {}
        self.parameter_checkboxes[model_id] = {}
        
        # Create the tab layout with better spacing
        tab_content = widgets.HBox(
            [fig, params_container],
            layout=widgets.Layout(
                justify_content='flex-start',
                padding='20px',
                height='600px',
            )
        )
        
        # Add to tabs
        current_tabs = list(self.tabs.children)
        current_tabs.append(tab_content)
        self.tabs.children = tuple(current_tabs)
        
        # Set the tab title
        self.tabs.set_title(len(current_tabs) - 1, model_name)
        
        # Store the model ID
        self.tab_ids.append(model_id)
        
        # Store q range controls
        self.parameter_controls[model_id]["q_min"] = q_min_input
        self.parameter_controls[model_id]["q_max"] = q_max_input
        
        # Store filter checkboxes
        self.parameter_checkboxes[model_id]["polydispersity_filter"] = polydispersity_checkbox
        self.parameter_checkboxes[model_id]["magnetic_filter"] = magnetic_checkbox
        
        # Add observers for filter checkboxes
        def update_parameter_visibility(change):
            show_polydispersity = polydispersity_checkbox.value
            show_magnetic = magnetic_checkbox.value
            
            # Get all parameter groups (skip title, filter box, and q range box)
            param_groups = list(params_container.children)[3:]
            
            for param_group in param_groups:
                if isinstance(param_group, widgets.VBox):
                    # Get parameter name from title
                    title_html = param_group.children[0].value
                    m = re.search(r'<b>(.*?)</b>', title_html)
                    if m:
                        param_name = m.group(1).strip()
                        
                        # Check if parameter should be visible
                        is_pd_param = '_pd' in param_name
                        is_magnetic_param = (param_name.startswith('up_') or 
                                           '_M0' in param_name or 
                                           '_mphi' in param_name or 
                                           '_mtheta' in param_name)
                        
                        # Update visibility
                        if (is_pd_param and not show_polydispersity) or \
                           (is_magnetic_param and not show_magnetic):
                            param_group.layout.display = 'none'
                        else:
                            param_group.layout.display = None
        
        polydispersity_checkbox.observe(update_parameter_visibility, names='value')
        magnetic_checkbox.observe(update_parameter_visibility, names='value')
    
    def remove_model_tab(self, tab_index: int) -> None:
        """Remove a model tab.
        
        Parameters
        ----------
        tab_index : int
            Index of the tab to remove
        """
        if 0 <= tab_index < len(self.tabs.children):
            # Get the model ID
            model_id = self.tab_ids[tab_index]
            
            # Remove the figure
            if model_id in self.figures:
                del self.figures[model_id]
            
            # Remove parameter controls
            if model_id in self.parameter_controls:
                del self.parameter_controls[model_id]
            
            if model_id in self.parameter_checkboxes:
                del self.parameter_checkboxes[model_id]
            
            # Remove the tab
            current_tabs = list(self.tabs.children)
            del current_tabs[tab_index]
            self.tabs.children = tuple(current_tabs)
            
            # Remove the ID
            del self.tab_ids[tab_index]
    
    def add_parameter_control(
        self, 
        model_id: str, 
        param_name: str, 
        value: Union[float, str],
        bounds: Optional[Tuple[float, float]] = None,
        use_autosas: bool = False,
        use_bounds: bool = False
    ) -> None:
        """Add a control for a parameter.
        
        Parameters
        ----------
        model_id : str
            ID of the model
        param_name : str
            Name of the parameter
        value : Union[float, str]
            Initial value for the parameter
        bounds : Optional[Tuple[float, float]]
            Optional bounds for the parameter
        use_autosas : bool
            Whether to enable AutoSAS for this parameter
        use_bounds : bool
            Whether to enable bounds for this parameter
        """
        # Get the tab index
        tab_index = self.tab_ids.index(model_id)
        
        # Get the parameters container (second item in the HBox)
        params_container = self.tabs.children[tab_index].children[1]
        
        # Check if parameter should be initially visible
        is_pd_param = '_pd' in param_name
        is_magnetic_param = (param_name.startswith('up_') or 
                           '_M0' in param_name or 
                           '_mphi' in param_name or 
                           '_mtheta' in param_name)
        
        show_polydispersity = self.parameter_checkboxes[model_id]["polydispersity_filter"].value
        show_magnetic = self.parameter_checkboxes[model_id]["magnetic_filter"].value
        
        # Determine initial display state
        initial_display = 'none' if ((is_pd_param and not show_polydispersity) or 
                                   (is_magnetic_param and not show_magnetic)) else None
        
        # Create the appropriate input widget based on parameter type
        if param_name.endswith('_pd_type'):
            # Create a dropdown for distribution type
            input_widget = widgets.Dropdown(
                value=value,
                options=bounds if bounds else ["gaussian", "rectangular", "schulz"],
                description="Type:",
                style={'description_width': 'initial'},
                layout=widgets.Layout(width='180px', margin='5px 0px 0px 10px')
            )
        elif param_name.endswith('_pd_n'):
            # Create an integer input for number of points
            input_widget = widgets.IntText(
                value=int(value),
                description="Npts:",
                style={'description_width': 'initial'},
                layout=widgets.Layout(width='180px', margin='5px 0px 0px 10px')
            )
        elif param_name.endswith('_pd_nsigma'):
            # Create a float input for number of sigmas
            input_widget = widgets.FloatText(
                value=float(value),
                description="Nsigmas:",
                style={'description_width': 'initial'},
                layout=widgets.Layout(width='180px', margin='5px 0px 0px 10px')
            )
        elif param_name.endswith('_pd'):
            # Create a float input for polydispersity width
            input_widget = widgets.FloatText(
                value=float(value),
                description="Width:",
                style={'description_width': 'initial'},
                layout=widgets.Layout(width='180px', margin='5px 0px 0px 10px')
            )
        else:
            # Create a standard float input for regular parameters
            input_widget = widgets.FloatText(
                value=float(value),
                description="Value:",
                style={'description_width': 'initial'},
                layout=widgets.Layout(width='180px', margin='5px 0px 0px 10px')
            )
        
        # Create a checkbox for AutoSAS (include in model_inputs)
        autosas_checkbox = widgets.Checkbox(
            value=use_autosas,
            description="AutoSAS",
            indent=False,
            layout=widgets.Layout(width='90px', margin='5px 0px 0px 10px')
        )
        
        # Create a checkbox for bounds/slider
        bounds_checkbox = widgets.Checkbox(
            value=use_bounds,
            description="Bounds",
            indent=False,
            layout=widgets.Layout(width='80px', margin='5px 0px 0px 10px')
        )
        
        # If bounds are enabled, create the slider
        if use_bounds and bounds is not None:
            # Create a row for the parameter with just the value display
            param_row = widgets.HBox([input_widget, autosas_checkbox, bounds_checkbox])
            
            # Create a title for the parameter
            param_title = widgets.HTML(f"<b>{param_name}</b>")
            
            # Create a parameter group
            param_group = widgets.VBox([
                param_title,
                param_row
            ], layout=widgets.Layout(
                margin='10px 0px',
                border='1px solid #eee',
                padding='0px 0px 10px 0px',
                min_height='60px',
                flex='none',
                display=initial_display
            ))
            
            # Add to parameters container
            current_params = list(params_container.children)
            current_params.append(param_group)
            params_container.children = tuple(current_params)
            
            # Store references to the controls
            self.parameter_controls[model_id][param_name] = input_widget
            self.parameter_checkboxes[model_id][param_name] = {
                "autosas": autosas_checkbox,
                "bounds": bounds_checkbox
            }
            
            # Now add the slider
            if len(param_group.children) == 2:
                if value > 0 and bounds[0] > 0:
                    slider_widget = widgets.FloatLogSlider(
                        value=value,
                        base=10,
                        min=np.log10(bounds[0]),
                        max=np.log10(bounds[1]),
                        description='Value',
                        continuous_update=True,
                        layout=widgets.Layout(width='300px')
                    )
                else:
                    slider_widget = widgets.FloatSlider(
                        value=value,
                        min=bounds[0],
                        max=bounds[1],
                        step=(bounds[1]-bounds[0])/100,
                        description='Value',
                        continuous_update=True,
                        layout=widgets.Layout(width='300px')
                    )

                slider_min = widgets.FloatText(
                    value=bounds[0],
                    description='Min:',
                    layout=widgets.Layout(width='200px')
                )
                slider_max = widgets.FloatText(
                    value=bounds[1],
                    description='Max:',
                    layout=widgets.Layout(width='200px')
                )

                slider_row = widgets.VBox([
                    slider_widget, slider_min, slider_max
                ], layout=widgets.Layout(margin='5px 0px'))

                new_children = list(param_group.children) + [slider_row]
                param_group.children = tuple(new_children)

                def update_label(change):
                    input_widget.value = change['new']
                slider_widget.observe(update_label, names='value')

                def update_slider_min(change):
                    if slider_widget.__class__.__name__ == 'FloatSlider':
                        slider_widget.min = change['new']
                    else:
                        slider_widget.min = np.log10(change['new']) if change['new'] > 0 else -3
                slider_min.observe(update_slider_min, names='value')

                def update_slider_max(change):
                    if slider_widget.__class__.__name__ == 'FloatSlider':
                        slider_widget.max = change['new']
                    else:
                        slider_widget.max = np.log10(change['new']) if change['new'] > 0 else 2
                slider_max.observe(update_slider_max, names='value')
        else:
            # Create a row for the parameter
            param_row = widgets.HBox([input_widget, autosas_checkbox, bounds_checkbox])
            
            # Create a title for the parameter
            param_title = widgets.HTML(f"<b>{param_name}</b>")
            
            # Create a parameter group with fixed minimum height to prevent shrinkage
            param_group = widgets.VBox([
                param_title,
                param_row
            ], layout=widgets.Layout(
                margin='10px 0px',
                border='1px solid #eee',
                padding='0px 0px 10px 0px',
                min_height='60px',
                flex='none',
                display=initial_display
            ))
            
            # Add to parameters container
            current_params = list(params_container.children)
            current_params.append(param_group)
            params_container.children = tuple(current_params)
            
            # Store references to the controls
            self.parameter_controls[model_id][param_name] = input_widget
            self.parameter_checkboxes[model_id][param_name] = {
                "autosas": autosas_checkbox,
                "bounds": bounds_checkbox
            }
    
    def get_parameter_control(self, model_id: str, param_name: str) -> widgets.Widget:
        """Get the control for a parameter.
        
        Parameters
        ----------
        model_id : str
            ID of the model
        param_name : str
            Name of the parameter
            
        Returns
        -------
        widgets.Widget
            The control widget
        """
        if model_id in self.parameter_controls and param_name in self.parameter_controls[model_id]:
            return self.parameter_controls[model_id][param_name]
        return None
    
    def get_parameter_checkbox(self, model_id: str, param_name: str) -> widgets.Checkbox:
        """Get the checkbox for a parameter."""
        if model_id in self.parameter_checkboxes and param_name in self.parameter_checkboxes[model_id]:
            # Default to returning the AutoSAS checkbox
            return self.parameter_checkboxes[model_id][param_name]["autosas"]
        return None
    
    def get_bounds_checkbox(self, model_id: str, param_name: str) -> widgets.Checkbox:
        """Get the bounds checkbox for a parameter."""
        if model_id in self.parameter_checkboxes and param_name in self.parameter_checkboxes[model_id]:
            return self.parameter_checkboxes[model_id][param_name]["bounds"]
        return None
    
    def replace_with_slider(self, model_id: str, param_name: str, value: float, bounds: Tuple[float, float]) -> None:
        # Get the tab index and parameters container
        tab_index = self.tab_ids.index(model_id)
        params_container = self.tabs.children[tab_index].children[1]

        # Iterate over parameter groups and extract parameter name from HTML title
        for i, param_group in enumerate(params_container.children):
            if isinstance(param_group, widgets.VBox):
                title_html = param_group.children[0].value
                m = re.search(r'<b>(.*?)</b>', title_html)
                if m and param_name.lower() in m.group(1).strip().lower():
                    # Ensure that the parameter group has only the title and controls row
                    if len(param_group.children) > 2:
                        param_group.children = param_group.children[:2]
                    controls_row = param_group.children[1]
                    if len(param_group.children) == 2:
                        if value > 0 and bounds[0] > 0:
                            slider_widget = widgets.FloatLogSlider(
                                value=value,
                                base=10,
                                min=np.log10(bounds[0]),
                                max=np.log10(bounds[1]),
                                description='Value',
                                #style={'description_width': 'initial'},
                                continuous_update=True,
                                layout=widgets.Layout(width='300px')
                            )
                        else:
                            slider_widget = widgets.FloatSlider(
                                value=value,
                                min=bounds[0],
                                max=bounds[1],
                                step=(bounds[1]-bounds[0])/100,
                                description='Value',
                                #style={'description_width': 'initial'},
                                continuous_update=True,
                                layout=widgets.Layout(width='300px')
                            )

                        slider_min = widgets.FloatText(
                            value=bounds[0],
                            description='Min:',
                            layout=widgets.Layout(width='200px')
                        )
                        slider_max = widgets.FloatText(
                            value=bounds[1],
                            description='Max:',
                            layout=widgets.Layout(width='200px')
                        )

                        slider_row = widgets.VBox([
                            slider_widget, slider_min, slider_max
                        ], layout=widgets.Layout(margin='5px 0px'))

                        new_children = list(param_group.children) + [slider_row]
                        param_group.children = tuple(new_children)

                        self.parameter_controls[model_id][param_name + '_slider'] = slider_widget

                        def update_label(change):
                            controls_row.children[0].value = change['new']
                        slider_widget.observe(update_label, names='value')

                        def update_slider_min(change):
                            if slider_widget.__class__.__name__ == 'FloatSlider':
                                slider_widget.min = change['new']
                            else:
                                slider_widget.min = np.log10(change['new']) if change['new'] > 0 else -3
                        slider_min.observe(update_slider_min, names='value')

                        def update_slider_max(change):
                            if slider_widget.__class__.__name__ == 'FloatSlider':
                                slider_widget.max = change['new']
                            else:
                                slider_widget.max = np.log10(change['new']) if change['new'] > 0 else 2
                        slider_max.observe(update_slider_max, names='value')
                    break
    
    def replace_with_text(self, model_id: str, param_name: str, value: float) -> None:
        tab_index = self.tab_ids.index(model_id)
        params_container = self.tabs.children[tab_index].children[1]

        for i, param_group in enumerate(params_container.children):
            if isinstance(param_group, widgets.VBox):
                title_html = param_group.children[0].value
                m = re.search(r'<b>(.*?)</b>', title_html)
                if m and param_name.lower() in m.group(1).strip().lower():
                    # Get the current checkboxes to preserve their state
                    current_controls = param_group.children[1]
                    autosas_checkbox = None
                    bounds_checkbox = None
                    
                    # Find the checkboxes in the current controls
                    for child in current_controls.children:
                        if isinstance(child, widgets.Checkbox):
                            if child.description == "AutoSAS":
                                autosas_checkbox = child
                            elif child.description == "Bounds":
                                bounds_checkbox = child
                    
                    # If we couldn't find the checkboxes, create new ones
                    if not autosas_checkbox:
                        autosas_checkbox = widgets.Checkbox(
                            value=False,
                            description="AutoSAS",
                            indent=False,
                            layout=widgets.Layout(width='90px', margin='5px 0px 0px 10px')
                        )
                    
                    if not bounds_checkbox:
                        bounds_checkbox = widgets.Checkbox(
                            value=False,
                            description="Bounds",
                            indent=False,
                            layout=widgets.Layout(width='80px', margin='5px 0px 0px 10px')
                        )
                    
                    # Create a new text input
                    text_input = widgets.FloatText(
                        value=value,
                        description="Value:",
                        style={'description_width': 'initial'},
                        layout=widgets.Layout(width='180px', margin='5px 0px 0px 10px')
                    )
                    
                    # Create a new controls row with all elements
                    new_controls_row = widgets.HBox([text_input, autosas_checkbox, bounds_checkbox])
                    
                    # Remove any slider row if it exists
                    param_group.children = (param_group.children[0], new_controls_row)
                    
                    # Update the control references
                    self.parameter_controls[model_id][param_name] = text_input
                    self.parameter_checkboxes[model_id][param_name] = {
                        "autosas": autosas_checkbox,
                        "bounds": bounds_checkbox
                    }
                    
                    # # Add observer for the text input
                    # def update_label(change):
                    #     text_input.value = f"{change['new']:.4g}"
                    # text_input.observe(update_label, names='value')
                    break
    
    def update_plot(self, model_id: str, q: np.ndarray, intensity: np.ndarray) -> None:
        """Update the plot for a model.
        
        Parameters
        ----------
        model_id : str
            ID of the model
        q : np.ndarray
            Array of q values
        intensity : np.ndarray
            Array of intensity values
        """
        if model_id in self.figures:
            fig = self.figures[model_id]
            with fig.batch_update():
                if len(q) > 0 and len(intensity) > 0:  # Only update if we have data
                    fig.data[0].x = q
                    fig.data[0].y = intensity
                
                # Update experimental data if available
                if self.data is not None and len(fig.data) > 1:
                    data_var = self.dropdown["data_var"].value
                    q_dim = self.dropdown["q_dim"].value
                    if data_var and q_dim:
                        da = self.data[data_var]
                        
                        # Get q values from the selected q dimension
                        q_data = da[q_dim].values if q_dim in da.dims else None
                        
                        # Get intensity data
                        if len(da.dims) == 1:
                            # 1D data
                            i_data = da.values
                        else:
                            # 2D data with index
                            non_q_dims = [dim for dim in da.dims if dim != q_dim]
                            if non_q_dims:
                                index_dim = non_q_dims[0]
                                index = int(self.text_input["data_index"].value)
                                i_data = da.isel({index_dim: index}).values
                            else:
                                i_data = None

                        if q_data is not None and i_data is not None:
                            fig.data[1].x = q_data
                            fig.data[1].y = i_data
                        else:
                            fig.data[1].x = []
                            fig.data[1].y = []
    
    def clear_tabs(self) -> None:
        """Clear all tabs."""
        # Clear tab IDs
        self.tab_ids = []
        
        # Clear parameter controls
        self.parameter_controls = {}
        self.parameter_checkboxes = {}
        
        # Clear figures
        self.figures = {}
        
        # Clear tabs
        self.tabs.children = ()

# Example usage in a Jupyter notebook
if __name__ == "__main__":
    # This code will run when the file is executed directly
    # Helpful for testing and demonstration
    
    print("""
    # Example of how to use AutoSASWidget in a notebook:
    
    from AFL.double_agent.AutoSASWidget import AutoSASWidget
    
    # Create the widget
    widget = AutoSASWidget()
    
    # Display it
    widget.run()
    
    # Later, after configuring models, you can get the model inputs:
    model_inputs = widget.get_model_inputs()
    
    # And use them with AutoSAS:
    # from AFL.double_agent.AutoSAS import AutoSAS
    # autosas = AutoSAS(..., model_inputs=model_inputs)
    """) 



class Labeler:
    def __init__(self,da):
        self.da = da
        self.fig,self.axes = plt.subplots(1,2,figsize=(8,4))
        AFL_Agent.format_plot_ternary(
            self.axes[1],
            label_b='PS',
            label_c='PS-PMMA',
            label_a='PMMA'
        )
        
        
        self.line_I =  self.da[0].plot(xscale='linear',yscale='log',ax=self.axes[0])[0]
        self.line_x =  self.axes[1].plot(0.5,0.5,color='red',marker='x')[0]
        self.axes[0].set(ylim=[None,1e4],xlim=(None,0.1))
        
        self.qstar = 0.01
        self.vlines = []
        for _ in range(3):
            l = self.axes[0].axvline(self.qstar,ls=':',alpha=0.5,lw=0.5,color='red')
            self.vlines.append(l)
        
        self.reset()
    
    def reset(self,*args,**kwargs):
        self.index = 0
        self.labels = []
        self.update_plot()
    
    def ternary_grid(self):
        others = [[1,2],[0,2],[0,1]]
        for i in [0,1,2]:
            for x in np.linspace(0,100,6):
                comp1 = np.zeros((2,3))
                comp1[:,i] = x
                comp1[0,others[i][0]] = 100-x
                comp1[1,others[i][1]] = 100-x
                cart = comp2cart(comp1)
                self.axes[1].plot(*cart.T,ls=':',alpha=0.5,lw=0.5,color='black')

    def update_plot(self):
        self.line_I.set_xdata(self.da[self.index].q.values)
        self.line_I.set_ydata(self.da[self.index].values)
    
        comp = np.array([self.da[self.index].concat_dim.item()])
        cart = comp2cart(comp*100.0)
        self.line_x.set_xdata(cart[0][0])
        self.line_x.set_ydata(cart[0][1])
        
        title = 'PS:{}, PS-PMMA:{}, PMMA:{}'.format(*comp[0])
        self.axes[0].set(title=title)
        
    def label(self,label):
        self.labels.append(label)
        self.index+=1
        self.update_plot()
        
    def label_C(self,*args,**kwargs):
        self.label('C')
        
    def label_L(self,*args,**kwargs):
        self.label('L')
        
    def label_S(self,*args,**kwargs):
        self.label('S')
        
    def label_D(self,*args,**kwargs):
        self.label('D')
        
    def update_peak_locations(self,*args,**kwargs):
        if self.phase_models.value=='LAM':
            locs = [1,2,3]
        elif self.phase_models.value=='HEX1':
            locs = [1,np.sqrt(3),np.sqrt(4)]
        elif self.phase_models.value=='HEX2':
            locs = [1,2/np.sqrt(3),np.sqrt(3)]
        elif self.phase_models.value=='SPH':
            locs = [1,np.sqrt(2),np.sqrt(3)]
        else:
            raise ValueError('Peak model not recognized')
        
        for line in self.vlines:
            line.set_linestyle('None')
        self.fig.canvas.draw()
        
        for loc,line in zip(locs,self.vlines):
            line.set_xdata([self.qstar*loc,self.qstar*loc])
            line.set_linestyle('-')
        self.fig.canvas.draw()
        
    def onclick(self,event):
        self.qstar=event.xdata
        self.update_peak_locations()
        # tx = 'button=%d, x=%d, y=%d, xdata=%f, ydata=%f' % (event.button, event.x, event.y, event.xdata, event.ydata)
        # self.axes[0].set_title(tx)
        # with self.output:
        #     print(tx)
    
    def show(self):
        
        self.b0 = ipywidgets.Button(description='Cylinder (yellow)')
        self.b1 = ipywidgets.Button(description='Lamellae (purple)')
        self.b2 = ipywidgets.Button(description='Sphere (red)')
        self.b3 = ipywidgets.Button(description='Disordered (blue)')
        self.br = ipywidgets.Button(description='Reset')
        
        self.phase_models = ipywidgets.Dropdown(
            options=['LAM','HEX1','HEX2','SPH']
        )
        self.phase_models.observe(self.update_peak_locations)
        
        hbox = ipywidgets.HBox([self.b0,self.b1,self.b2,self.b3,self.br])
        self.output = ipywidgets.Output()
        vbox = ipywidgets.VBox([hbox,self.phase_models,self.output])
        
        self.b0.on_click(self.label_C)
        self.b1.on_click(self.label_L)
        self.b2.on_click(self.label_S)
        self.b3.on_click(self.label_D)
        self.br.on_click(self.reset)
            
        self.fig.canvas.mpl_connect('button_press_event',self.onclick)
        #self.axes[0].callbacks.connect('button_press_event',self.onclick)
        
        return vbox
        


import pickle
import warnings
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from sklearn.preprocessing import OrdinalEncoder


class PhaseMap:
    '''Container for compositions, measurements, and labels'''
    def __init__(self,components):
        self.model = PhaseMapModel(components)
        self.view = PhaseMapView_MPL()
        
    def __str__(self):
        return f'<PhaseMap {self.shape[0]} pts>'
    
    def __repr__(self):
        return self.__str__()
            
    def __getitem__(self,index):
        composition = self.model.compositions.iloc[index]
        measurement = self.model.measurements.iloc[index]
        label = self.model.labels.iloc[index]
        return (composition,measurement,label)
    
    def slice(self,indices):
        compositions = self.model.compositions.iloc[indices]
        measurements = self.model.measurements.iloc[indices]
        labels = self.model.labels.iloc[indices]
        pm = PhaseMap(self.components)
        pm.append(
            compositions=compositions,
            measurements=measurements,
            labels=labels,
        )
        return pm
    
    def project(self,components):
        compositions = self.compositions[components].copy()
        compositions,mask = rescale_compositions(compositions)
        
        labels = self.labels.loc[mask].copy()
        measurements = self.measurements.reset_index(drop=True).loc[mask].copy()
        
        pm = PhaseMap(components)
        pm.append(
            compositions=compositions,
            measurements=measurements,
            labels=labels,
        )
        return pm
        

    @property
    def components(self):
        return self.model.components

    @property
    def ncomponents(self):
        return len(self.model.components)
    
    @property
    def compositions(self):
        return self.model.compositions
    
    @compositions.setter
    def compositions(self,value):
        self.model.compositions = value
    
    @property
    def measurements(self):
        return self.model.measurements

    @measurements.setter
    def measurements(self,measurements):
        if not isinstance(measurements,pd.DataFrame):
            measurements = pd.DataFrame(measurements.copy())
        self.model.measurements = measurements
    
    @property
    def labels(self):
        return self.model.labels
    
    @labels.setter
    def labels(self,labels):
        if not isinstance(labels,pd.Series):
            labels = pd.Series(labels.copy())
        self.model.labels = labels.copy()
        
    @property
    def labels_ordinal(self):
        '''Numerical labels sorted by spatial position'''
        self.update_encoder()
        
        labels_ordinal = pd.Series(
            data=self.label_encoder.transform(
                self.labels.values.reshape(-1,1)
            ).flatten(),
            index=self.model.labels.index,
        )
        return labels_ordinal
    
    def update_encoder(self):
        labels_sorted = (
            self.compositions
            .copy()
            .set_index(self.labels)
            .sort_values(self.components[:-1]) # sort by comp1 and then comp2
            .index
            .values
        )
        
        self.label_encoder.fit(labels_sorted.reshape(-1,1))
    
    @property
    def label_encoder(self):
        return self.model.label_encoder
    
    @property
    def shape(self):
        return self.model.compositions.shape
    
    @property
    def size(self):
        return self.model.compositions.size
    
    def copy(self,labels=None):
        if labels is None:
            labels = self.model.labels
           
        if not isinstance(labels,pd.Series):
            labels = pd.Series(labels)
            
        pm = PhaseMap(self.components)
        pm.append(
            compositions = self.model.compositions,
            measurements = self.model.measurements,
            labels = labels,
        )
        return pm
    
    def save(self,fname):
        out_dict = {}
        out_dict['components'] = self.model.components
        out_dict['compositions'] = self.model.compositions
        out_dict['measurements'] = self.model.measurements
        out_dict['labels'] = self.model.labels
        
        fname = str(fname)#handle pathlib objects
        if not (fname[-4:]=='.pkl'):
            fname+='.pkl'
        
        with open(fname,'wb') as f:
            pickle.dump(out_dict,f,protocol=-1)
            
    @classmethod
    def load(cls,fname):
        fname = str(fname)#handle pathlib objects
        if not (fname[-4:]=='.pkl'):
            fname+='.pkl'
            
        with open(fname,'rb') as f:
            in_dict = pickle.load(f)
        
        pm = cls(in_dict['components'])
        pm.append(
            compositions = in_dict['compositions'],
            measurements = in_dict['measurements'],
            labels       = in_dict['labels'],
        )
        return pm
            
    
    def append(self,compositions,measurements=None,labels=None):
        self.model.append(
                compositions=compositions,
                measurements=measurements,
                labels=labels
                )
        
    def plot(self,components=None,compositions=None,labels=None,rescale=True,**mpl_kw):
        if (components is None) and (compositions is None) and (labels is None):
            components = self.components.copy()
            labels = self.labels.copy()
            compositions = self.compositions.copy()
            
        if components is not None:
            compositions = self.compositions[components].copy()
            labels = self.labels.copy()
        else:
            components = self.components.copy()
            
        if labels is None:
            labels = pd.Series(np.ones_like(compositions.shape[0]))
        
        if rescale:
            # compositions = compositions.apply(lambda x: 100.0*x/x.sum(),axis=1)
            comps = compositions.copy()
            mask = (comps[components].sum(1)>0.0)
            comps = comps[components].loc[mask].values
            comps = 100.0*comps/comps.sum(1)[:,np.newaxis]
            compositions = pd.DataFrame(comps,columns=components)
            labels = labels.loc[mask]
        
        compositions = compositions[components]#ensure ordering
            
        ax = None
        if len(components)==2:
            ax=self.view.scatter(compositions,labels=labels,**mpl_kw)
        elif len(components)==3:
            ax=self.view.scatter_ternary(compositions,labels=labels,**mpl_kw)
        else:
            raise ValueError(f'Unable to plot {len(components)} dimensions.')
        return ax
    
class PhaseMapModel:
    def __init__(self,components):
        self.components = components
        self.compositions = pd.DataFrame(columns=components)
        self.measurements = pd.DataFrame()
        self.labels = pd.Series(dtype=np.float64)
        self.label_encoder = OrdinalEncoder()
        
    def append(self,compositions,labels,measurements):
        self.compositions = pd.concat([self.compositions,compositions],ignore_index=True)
        if measurements is not None:
            self.measurements = pd.concat([self.measurements,measurements])
        else:
            self.measurements = None
            
        if labels is not None:
            self.labels = pd.concat([self.labels,labels],ignore_index=True)
        else:
            self.labels = pd.Series(np.ones(compositions.shape[0]))
        
    
class PhaseMapView_MPL:
    def __init__(self,cmap='jet'):
        self.cmap = cmap
        
    def make_axes(self,components,subplots=(1,1)):
        fig,ax = plt.subplots(*subplots,figsize=(5*subplots[1],4*subplots[0]))
        
        if (subplots[0]>1) or (subplots[1]>1):
            ax = ax.flatten()
            if len(components)==3:
                for cax in ax:
                    format_plot_ternary(ax,*components)
        else:
            if len(components)==3:
                format_plot_ternary(ax,*components)
        return ax
    
    def scatter(self,compositions,labels=None,ax=None,**mpl_kw):
        if ax is None:
            ax = self.make_axes(compositions.columns.values,(1,1))
        
        xy = compositions.values
        
        if 'marker' not in mpl_kw:
            mpl_kw['marker'] = '.'
        if ('cmap' not in mpl_kw) and ('color' not in mpl_kw):
            mpl_kw['cmap'] = self.cmap
        mpl_kw['c'] = labels
        ax.scatter(*xy.T,**mpl_kw)
        return ax
    
    def scatter_ternary(self,compositions,labels=None,ax=None,**mpl_kw):
        if ax is None:
            ax = self.make_axes(compositions.columns.values,(1,1))
        
        xy = ternary2cart(compositions.values)
        
        if 'marker' not in mpl_kw:
            mpl_kw['marker'] = '.'
        if ('cmap' not in mpl_kw) and ('color' not in mpl_kw):
            mpl_kw['cmap'] = self.cmap
        if 'color' not in mpl_kw:
            mpl_kw['c'] = labels
        ax.scatter(*xy.T,**mpl_kw)
        return ax
    
    def lines(self,xy,ax=None,label=None):
        if ax is None:
            ax = self.make_axes((1,1))
            
        ax.plot(*xy.T,marker='None',ls=':',label=label)
        return ax

def rescale_compositions(compositions,basis=100.0,remove_zeros=True):
    comps = compositions.copy()
    if remove_zeros:
        mask = (comps.sum(1)>0.0)
    else:
        mask = np.ones(comps.shape[0],dtype=bool)
        
    comps = comps.loc[mask].values
    comps = basis*comps/comps.sum(1)[:,np.newaxis]
    compositions = pd.DataFrame(comps,columns=compositions.columns)
    return compositions,mask

def phasemap_grid_factory(components,pts_per_row=50,basis=100):
    compositions = composition_grid(
            pts_per_row = pts_per_row,
            basis = basis,
            dim=len(components),
            )
    N = compositions.shape[0]
    compositions = pd.DataFrame(compositions,columns=components)
    q = np.geomspace(1e-3,1,25)
    I = np.random.random(25)
    measurements = pd.concat([pd.Series(index=q,data=I) for _ in range(N)],axis=1).T
    labels = pd.Series(np.ones(N))
     
    pm = PhaseMap(components)
    pm.append(
            compositions=compositions,
            measurements=measurements,
            labels=labels,
            )

    return pm


def composition_grid(pts_per_row=50,basis=100,dim=3,eps=1e-9):
    try:
        from tqdm.contrib.itertools import product
    except ImportError:
        from itertools import product
    pts = []
    for i in product(*[np.linspace(0,1.0,pts_per_row)]*(dim-1)):
        if sum(i)>(1.0+eps):
            continue
            
        j = 1.0-sum(i)
        
        if j<(0.0-eps):
            continue
        pt = [k*basis for k in [*i,j]]
        pts.append(pt)
    return np.array(pts)

def cart2ternary(compositions):
    '''Ternary composition to Cartesian cooridate'''
    if compositions is None:
        compositions = self.model.compositions
        
    try:
        #Assume pandas
        xy = compositions.values.copy()
    except AttributeError:
        # Assume numpy
        xy = compositions.copy()
    
    if xy.ndim==1:
        xy = np.array([xy])

    if xy.shape[1]!=2:
        raise ValueError('This class only works with ternary (3-component) data!')
        
    # Convert ternary data to cartesian coordinates.
    t = np.zeros((xy.shape[0],3))
    t[:,1] = 100.*xy[:,1]/np.sin(60*np.pi/180.)
    t[:,0] = 100.0*(xy[:,0] - xy[:,1]*np.sin(30.*np.pi/180.)/np.sin(60*np.pi/180))
    t[:,2] = 100.0 - t[:,0] - t[:,1]
    return t

def ternary2cart(compositions):
    '''Ternary composition to Cartesian cooridate'''
    if compositions is None:
        compositions = self.model.compositions
        
    try:
        #Assume pandas
        t = compositions.values.copy()
    except AttributeError:
        # Assume numpy
        t = compositions.copy()
    
    if t.ndim==1:
        t = np.array([t])

    if t.shape[1]!=3:
        raise ValueError('This class only works with ternary (3-component) data!')
        
    # Convert ternary data to cartesian coordinates.
    xy = np.zeros((t.shape[0],2))
    xy[:,1] = t[:,1]*np.sin(60.*np.pi / 180.) / 100.
    xy[:,0] = t[:,0]/100. + xy[:,1]*np.sin(30.*np.pi/180.)/np.sin(60*np.pi/180)
    return xy

def format_plot_ternary(ax,label_a=None,label_b=None,label_c=None):
    ax.axis('off')
    ax.set(
        xlim = [0,1],
        ylim = [0,1],
    )
    ax.plot([0,1,0.5,0],[0,0,np.sqrt(3)/2,0],ls='-',color='k')
    if label_a is not None:
        ax.text(1,0,label_a,ha='left')
    if label_b is not None:
        ax.text(0.5,np.sqrt(3)/2,label_b,ha='center',va='bottom')
    if label_c is not None:
        ax.text(0,0,label_c,ha='right')



"""
Datasets module for AFL.double_agent.

This module provides access to example datasets that can be used with AFL.double_agent.
"""

import os
import pathlib
import warnings

import xarray as xr


# Get the path to the data directory
def get_data_dir() -> pathlib.Path:
    """
    Get the path to the data directory.

    Returns
    -------
    pathlib.Path
        Path to the data directory.
    """
    # The data directory is now located in AFL/double_agent/data
    module_dir = pathlib.Path(__file__).parent
    data_dir = module_dir

    if not data_dir.exists():
        warnings.warn(f"Data directory not found at {data_dir}")

    return data_dir


def list_datasets() -> list[str]:
    """
    List all available datasets.

    Returns
    -------
    list
        List of available dataset names.
    """
    data_dir = get_data_dir()
    if not data_dir.exists():
        warnings.warn(f"Data directory not found at {data_dir}")
        return []

    return [f.stem for f in data_dir.glob("*.nc")]


def load_dataset(name: str, lazy: bool=False) -> xr.Dataset:

    """
    Load a dataset by name.

    Parameters
    ----------
    name : str
        Name of the dataset to load.

    Returns
    -------
    xarray.Dataset
        The loaded dataset.

    Raises
    ------
    FileNotFoundError
        If the dataset does not exist.
    """
    data_dir = get_data_dir()
    file_path = data_dir / f"{name}.nc"

    if not file_path.exists():
        raise FileNotFoundError(
            f"Dataset '{name}' not found at {file_path}. "
            f"Data directory: {data_dir}. "
            f"Available datasets: {list_datasets()}"
        )
        
    if lazy:
        return xr.open_dataset(file_path)
    else:
        return xr.load_dataset(file_path)


# Define specific dataset loaders
def example_dataset1() -> xr.Dataset:
    """
    Load the example dataset.

    Returns
    -------
    xarray.Dataset
        The example dataset.
    """
    return load_dataset("example_dataset")

  
# Define specific dataset loaders
def example_dataset2():
    """
    Load the example dataset.
    
    Returns
    -------
    xarray.Dataset
        The example dataset.
    """
    return load_dataset("synthetic_sans")

# Add all datasets as module-level variables
__all__ = ["load_dataset", "list_datasets", "example_dataset1", "example_dataset2"] 


from AFL.automation.APIServer.Client import Client
from AFL.automation.prepare.OT2Client import OT2Client
from AFL.automation.shared.utilities import listify
from AFL.automation.APIServer.Driver import Driver
from AFL.agent.AgentClient import AgentClient
from AFL.automation.shared.units import units

from AFL.agent.reduce_usaxs import reduce_uascan

from scipy.spatial.distance import cdist

import warnings

import pandas as pd
import numpy as np 
import xarray as xr 
from math import ceil,sqrt
import json
import time
import requests
import shutil
import datetime
import traceback
import pathlib
import uuid
import copy

import AFL.automation.prepare
import shutil
import h5py


class SAS_Grid_AL_SampleDriver(Driver):
    defaults={}
    defaults['data_path'] = '/mnt/aux/chess/reduced_data/cycles/2022-1/id3b/beaucage-2324-D/analysis/'
    defaults['data_manifest_file'] = '/home/afl642/USAXS_data/usaxs_manifest.csv'
    defaults['AL_manifest_file'] = '/home/afl642/USAXS_data/AL_manifest.nc'
    defaults['sample_manifest_file'] = '/home/afl642/USAXS_data/sample_manifest.nc'
    defaults['data_tag'] = 'default'
    def __init__(self,
            sas_url,
            agent_url,
            overrides=None, 
            ):

        Driver.__init__(self,name='_AL_SampleDriver',defaults=self.gather_defaults(),overrides=overrides)

        if not (len(agent_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on agent_url')

        self.app = None
        self.name = 'SAS_Grid_AL_SampleDriver'

        #measure samples
        self.sas_url = sas_url
        self.sas_client = Client(sas_url.split(':')[0],port=sas_url.split(':')[1])
        self.sas_client.login('SampleServer_SASClient')
        self.sas_client.debug(False)
        
        #agent
        self.agent_url = agent_url
        self.agent_client = AgentClient(agent_url.split(':')[0],port=agent_url.split(':')[1])
        self.agent_client.login('SampleServer_AgentClient')
        self.agent_client.debug(False)

        self.status_str = 'Fresh Server!'
        
        self.catch_protocol=None
        self.AL_status_str = ''
        self.sample_manifest  = None
        self.data_manifest  = None
        self.AL_manifest  = None
        self.AL_components = None
        self.num_samples = 0
        self.blank_iter = None
        self.blank_well = None
        self.blank_plate_name = None
        
        self.loaded_plates = {}
        
    def status(self):
        status = []
        status.append(f'SAS: {self.sas_url}')
        status.append(self.status_str)
        status.append(self.AL_status_str)
        status.append(f'AL Components: {self.AL_components}')
        status.append(f'{self.num_samples} samples loaded')
        status.append(f'Measuring blank well {self.blank_well} on plate {self.blank_plate_name} every {self.blank_iter} iterations')
        return status
    
    def update_status(self,value):
        self.status_str = value
        self.app.logger.info(value)

    def measure(self,sample):
        exposure = sample.get('exposure',None)

        sas_uuid = self.sas_client.enqueue(task_name='expose',name=sample['name'],block=True,exposure=exposure)
        self.sas_client.wait(sas_uuid)

        return sas_uuid

    def sample(self,sample):
        return self.process_sample(sample)
    
    @Driver.unqueued()
    def stop_active_learning(self):
        self.stop_AL = True
        
        
    def active_learning_loop(self,**kwargs):
        self.AL_selection  = kwargs['AL_selection']
        pre_run_list       =  copy.deepcopy(kwargs.get('pre_run_list',[]))
        plate_to_slot      = kwargs['plate_to_slot']
        predict            = kwargs.get('predict',True)
        start_new_manifest            = kwargs.get('start_new_manifest',False)
        pts_per_row            = kwargs.get('pts_per_row',25)

        #handle blanks
        self.blank_iter            = kwargs.get('blank_iter',None)
        self.blank_well            = kwargs.get('blank_well',None)
        self.blank_plate_name      = kwargs.get('blank_plate_name',None)
        if self.blank_plate_name is not None:
            blank_plate_num = self.blank_plate_name.split('_')[1]
        if self.blank_well is not None:
            blank_row = self.blank_well[0]
            blank_col = int(self.blank_well[1:])


        # grab paths from config
        AL_manifest_path = pathlib.Path(self.config['AL_manifest_file'])
        sample_manifest_path = pathlib.Path(self.config['sample_manifest_file'])
        data_path = pathlib.Path(self.config['data_path'])
        data_manifest_path = pathlib.Path(self.config['data_manifest_file'])
        
        ################################
        ## LOAD MANIFEST & DOWNSELECT ##
        ################################
        self.sample_manifest = xr.load_dataset(sample_manifest_path)
        self.AL_selection['plate_name'] = list(plate_to_slot.keys())
        self.sample_manifest = mask_dataset(self.sample_manifest,self.AL_selection)
        self.num_samples = self.sample_manifest.sizes['sample']

        self.AL_components = self.sample_manifest.attrs['components']
        
        
        AL_iter = -1
        self.stop_AL = False
        while not self.stop_AL:
            AL_iter+=1

            self.app.logger.info(f'Starting new AL loop')

            #############################
            ## PERIODIC MEASURE BLANKS ##
            #############################
            if (AL_iter>0) and (self.blank_iter is not None) and (self.blank_well is not None) and (self.blank_plate_name is not None) and ((AL_iter%self.blank_iter)==0):
                sas_fname = f'p{blank_plate_num}_{self.blank_well}_y2_blank'
                blank_slot = plate_to_slot[self.blank_plate_name]
                self.update_status(f'Measuring blank well {self.blank_well} on {self.blank_plate_name} in slot {blank_slot}')
                self.sas_client.enqueue(task_name='setPosition',plate=blank_slot,row=blank_row,col=blank_col,y_offset=2)
                sas_uuid = self.sas_client.enqueue(task_name='expose',name=sas_fname,block=True)
                # self.sas_client.wait(sas_uuid)
            
            ###########################
            ## GET NEXT STEP FROM AL ##
            ###########################
            self.update_status('Getting next step from AL...')
            if pre_run_list:
                self.next_sample_AL = None
                next_sample_dict = pre_run_list.pop(0)
            else:
                #next_sample = self.agent_client.get_next_sample_queued()
                self.next_sample_AL = self.agent_client.get_object('next_sample')
                try:
                    next_sample_dict = self.next_sample_AL.squeeze().reset_coords('grid',drop=True).to_pandas().to_dict()
                except ValueError:
                    next_sample_dict = self.next_sample_AL.squeeze().to_pandas().to_dict()
                
            self.app.logger.info(f'Trying to measure next sample: {next_sample_dict}')
            
            ##############################
            ## FIND CLOSEST IN MANIFEST ##
            ##############################
            self.update_status('Finding closest sample in manifest...')
            
            # find closest in manifest
            coords_available = self.sample_manifest[self.AL_components].to_array('component')
            ds_next_sample = pd.Series(next_sample_dict).to_frame().T.to_xarray().rename_dims(index='sample').reset_coords()
            coords_new = ds_next_sample[self.AL_components].to_array('component').transpose(...,'component')
    
            coords_available = self.sample_manifest[self.AL_components].to_array('component').transpose(...,'component')
            sample_distances = cdist(coords_new,coords_available)
            self.next_sample = self.sample_manifest.isel(sample=sample_distances.argmin()).reset_coords()
            self.sample_manifest = self.sample_manifest.drop_isel(sample=sample_distances.argmin())#remove entry from sample_list
            self.num_samples = self.sample_manifest.sizes['sample']#update num samples
            
            ###############################
            ## MEASURE OR READ FROM DISK ##
            ###############################
            # check if already measured
            next_plate_name = self.next_sample.plate_name.values[()]
            next_plate_num = int(self.next_sample.plate.values[()])
            _,next_well = parse_well(self.next_sample.dest.values[()])
            next_well_row = next_well[0]
            next_well_col = int(next_well[1:])
            sas_fname = f'p{next_plate_num}_{next_well}_y2'
            slot_name = plate_to_slot[next_plate_name]

            # reload data manifest and check if already measured
            self.df_data_manifest = pd.read_csv(data_manifest_path)
            try:
                sel = self.df_data_manifest.set_index(['plate','well']).loc[next_plate_num,next_well]
            except KeyError:
                sel = None

            
            if sel is None:
                self.update_status(f'Measuring well {next_well} on plate {next_plate_name} in slot {slot_name}')
                self.sas_client.enqueue(task_name='setPosition',plate=slot_name,row=next_well_row,col=next_well_col,y_offset=2)
                sas_uuid = self.sas_client.enqueue(task_name='expose',name=sas_fname,block=True)
                self.sas_client.wait(sas_uuid)

                sleep_time=5
                for i in range(25):
                    print(f'--> Iteration {i} on trying to find sas data...waited {i*sleep_time} seconds...')
                    sas_fpath = list(data_path.glob(f'{sas_fname}*.h5'))
                    sas_fpath = sorted(sas_fpath,key = lambda x: str(x).split('_')[-1])
                    if len(sas_fpath)==0:
                        time.sleep(sleep_time)
                    else:
                        break

                if len(sas_fpath)==0:
                    raise ValueError('No file found after measurement...')
                else:
                    sas_fpath = sas_fpath[-1]


                # add to manifest
                row = {}
                row['well'] = next_well
                row['plate'] = next_plate_num
                row['plate_name'] = next_plate_name
                row['sample_name'] = self.next_sample.sample_name.values[()]
                row['sample'] = self.df_data_manifest.shape[0]
                row['MOF'] = self.next_sample.MOF.values[()]
                row['fpath'] = sas_fpath
                self.df_data_manifest  = self.df_data_manifest.append(row,ignore_index=True)
                self.df_data_manifest.to_csv(data_manifest_path)

            else:
                sas_fpath = sel.fpath
                sas_fpath = sorted(sel.fpath.values,key = lambda x: str(x).split('_')[-1])[-1]
                self.update_status(f'Reading data for well {next_well} on plate {next_plate_name} in slot {slot_name}: {sas_fpath}')
                
                
            #read data and add to next_sample dict
            self.new_data = self.next_sample.copy()
            with h5py.File(sas_fpath,'r') as h5:
                reduced = reduce_uascan(h5)

            da_sas = xr.DataArray(reduced['R'],coords={'q':reduced['Q']},dims=['q'])
            da_sas['q'] = da_sas.q.pipe(np.abs)
            da_sas = da_sas.groupby('q').mean()
            da_sas = da_sas.interp(q=np.geomspace(1e-5,1e-1,500))
            self.new_data['R'] = da_sas
            self.new_data.attrs['AL_data'] = 'R'
            
            ########################
            ## UPDATE AL MANIFEST ##
            ########################
            if start_new_manifest and AL_manifest_path.exists():
                self.update_status(f'Backing up old and then starting new manifest...')
                AL_manifest_backup = xr.load_dataset(AL_manifest_path)
                i=0
                while True:
                    AL_manifest_backup_path = pathlib.Path(str(AL_manifest_path)+f'.bak{i}')
                    if AL_manifest_backup_path.exists():
                        i+=1
                    else:
                        print(f'--> Writing backup AL file to {AL_manifest_backup_path}')
                        AL_manifest_backup.to_netcdf(AL_manifest_backup_path)
                        break

                self.AL_manifest = self.new_data.expand_dims('sample')
                start_new_manifest=False

            elif AL_manifest_path.exists():
                self.update_status(f'Updating manifest...')
                self.AL_manifest = xr.load_dataset(AL_manifest_path)

                if 'grid' in self.AL_manifest.dims:
                    self.AL_manifest = self.AL_manifest.drop_dims('grid')

                self.AL_manifest = xr.concat([self.AL_manifest,self.new_data],dim='sample')

            else:
                self.update_status(f'Starting new manifest...')
                self.AL_manifest = self.new_data.expand_dims('sample')
            
            self.AL_manifest.attrs.update(self.sample_manifest.attrs)
            self.AL_manifest.attrs['AL_data'] = 'R'

            self.AL_manifest  = self.AL_manifest.afl.comp.add_grid(pts_per_row=pts_per_row)

            # self.AL_manifest.attrs['components'] = self.AL_components
            # self.AL_manifest.attrs['GP_domain_transform'] = 'range_scaled'
            #  for component in self.AL_components:
            #      self.AL_manifest.attrs[f'{component}_range'] = self.AL_component_ranges[component]
            #      self.AL_manifest.attrs[f'{component}_grid_range'] = self.AL_component_ranges[component]
            self.AL_manifest.to_netcdf(AL_manifest_path)

            ################
            ## TRIGGER AL ##
            ################
            self.update_status(f'Triggering agent server...')
            if len(pre_run_list)==0:
                if predict:
                    self.agent_uuid = self.agent_client.enqueue(task_name='predict',datatype='nc')
                
                    # wait for AL
                    self.app.logger.info(f'Waiting for agent...')
                    self.agent_client.wait(self.agent_uuid)
                else:#used for intialization
                    return
            
            
   

def mask_dataset(dataset,sel_dict): 
    masks_and = []
    for var,values in sel_dict.items():
        masks_or = []
        for value in values:
            masks_or.append(dataset[var]==value)
        masks_and.append(np.logical_or.reduce(masks_or))
    mask = np.logical_and.reduce(masks_and)
    mask = xr.ones_like(dataset[var]).copy(data=mask)#just use last var in list...hopefully this isn't folly
    return dataset.copy().where(mask,drop=True)


   

def parse_well(loc):
    for i,loc_part in enumerate(list(loc)):
        if loc_part.isalpha():
            break
    slot = loc[:i]
    well = loc[i:]
    return slot,well
